"""Database operations for immichporter."""

from sqlalchemy.orm import Session
from sqlalchemy import func, or_, case
from typing import List, Dict, Any
from rich.console import Console
from loguru import logger

from immichporter.models import Base, Album, User, Photo, Error, AlbumUser, SessionLocal

from immichporter.schemas import AlbumInfo, PictureInfo

console = Console()

# Module-level variable to track if database has been initialized
_database_initialized = False


def get_db_session() -> Session:
    """Get a database session. Initializes and migrates the database if needed."""
    global _database_initialized
    if not _database_initialized:
        # Initialize the database only once per session
        init_database()
        _database_initialized = True
    return SessionLocal()


def init_database(reset_db: bool = False) -> None:
    """Initialize the database and apply any necessary migrations."""
    from sqlalchemy import inspect, text

    engine = SessionLocal().bind
    assert engine is not None

    if reset_db:
        Base.metadata.drop_all(bind=engine)

    # Create all tables if they don't exist
    Base.metadata.create_all(bind=engine)

    # Apply migrations
    inspector = inspect(engine)

    # Migration 1: Add add_to_immich column if it doesn't exist
    if "users" in inspector.get_table_names():
        columns = [col["name"] for col in inspector.get_columns("users")]
        if "add_to_immich" not in columns:
            with engine.begin() as conn:
                conn.execute(  # type: ignore
                    text(
                        "ALTER TABLE users ADD COLUMN add_to_immich BOOLEAN DEFAULT TRUE NOT NULL"
                    )
                )
                console.print(
                    "[yellow]Applied migration: Added add_to_immich column to users table[/yellow]"
                )

        # Migration 2: Add immich_user_id column if it doesn't exist
        if "immich_user_id" not in columns:
            with engine.begin() as conn:
                conn.execute(  # type: ignore
                    text("ALTER TABLE users ADD COLUMN immich_user_id INTEGER")
                )
                console.print(
                    "[yellow]Applied migration: Added immich_user_id column to users table[/yellow]"
                )
        if "immich_initial_password" not in columns:
            with engine.begin() as conn:
                conn.execute(  # type: ignore
                    text("ALTER TABLE users ADD COLUMN immich_initial_password STRING")
                )

    if "photos" in inspector.get_table_names():
        columns = [col["name"] for col in inspector.get_columns("photos")]
        if "immich_id" not in columns:
            with engine.begin() as conn:
                conn.execute(text("ALTER TABLE photos ADD COLUMN immich_id STRING"))  # type: ignore
                console.print(
                    "[yellow]Applied migration: Added immich_id column to photos table[/yellow]"
                )
        if "saved_to_your_photos" not in columns:
            with engine.begin() as conn:
                conn.execute(  # type: ignore
                    text(
                        "ALTER TABLE photos ADD COLUMN saved_to_your_photos BOOLEAN DEFAULT FALSE NOT NULL"
                    )
                )
                conn.execute(  # type: ignore
                    text(
                        "UPDATE photos SET saved_to_your_photos = FALSE WHERE saved_to_your_photos IS NULL"
                    )
                )
                console.print(
                    "[yellow]Applied migration: Added saved_to_your_photos column to photos table[/yellow]"
                )

    # Migration: Add immich_id to albums table
    if "albums" in inspector.get_table_names():
        columns = [col["name"] for col in inspector.get_columns("albums")]
        if "immich_id" not in columns:
            with engine.begin() as conn:
                conn.execute(text("ALTER TABLE albums ADD COLUMN immich_id STRING"))  # type: ignore
                console.print(
                    "[yellow]Applied migration: Added immich_id column to albums table[/yellow]"
                )

    logger.debug("Database initialized and migrated successfully")


def insert_or_update_album(session: Session, album_info) -> int:
    """Insert or update an album from AlbumInfo object."""
    # Check if album exists
    existing_album: Album | None = (
        session.query(Album)
        .filter_by(source_url=album_info.url, source_type="gphoto")
        .first()
    )

    if existing_album is not None:
        assert isinstance(existing_album, Album)
        # Update existing album
        existing_album.items = album_info.items
        existing_album.shared = album_info.shared
        existing_album.source_url = album_info.url
        session.commit()
        return existing_album.id
    else:
        # Insert new album
        album = Album(
            source_title=album_info.title,
            source_type="gphoto",
            immich_title=album_info.title,
            items=album_info.items,
            shared=album_info.shared,
            source_url=album_info.url,
        )
        session.add(album)
        session.commit()
        return album.id


def insert_or_update_user(
    session: Session, user_name: str, add_to_immich: bool = True
) -> int:
    """Insert or update a user.

    Args:
        session: Database session
        user_name: Name of the user in the source system
        add_to_immich: Whether to include this user in Immich imports (default: True)
    """
    # Check if user exists
    existing_user = (
        session.query(User)
        .filter_by(source_name=user_name, source_type="gphoto")
        .first()
    )

    if existing_user:
        logger.debug(f"User already exists: {user_name}")
        return existing_user.id
    else:
        # Insert new user
        user = User(
            source_name=user_name, source_type="gphoto", add_to_immich=add_to_immich
        )
        session.add(user)
        session.commit()
        logger.info(f"Added user: {user_name} (add_to_immich={add_to_immich})")
        return user.id


def insert_photo(
    session: Session, picture_info: PictureInfo, album_id: int, update: bool = False
) -> tuple[bool | None, int | None]:
    """Insert a photo, returning a tuple (updated, photo_id | None), updated is None if photo was not inserted.

    Returns:
        A tuple (update, photo_id | None)
    """
    logger.debug(f"Inserting photo: {picture_info}")
    # Check if photo already exists
    existing_photo = (
        session.query(Photo)
        .filter_by(source_id=picture_info.source_id, album_id=album_id)
        .first()
    )

    if existing_photo and not update:
        logger.info(f"Photo already exists: {picture_info.filename} (not updated)")
        return None, existing_photo.id
    elif existing_photo:  # update
        something_changed = False
        for attr in [
            "filename",
            "date_taken",
            "user_id",
            "source_id",
            "saved_to_your_photos",
        ]:
            if getattr(picture_info, attr) != getattr(existing_photo, attr):
                if something_changed is False:
                    logger.info(
                        f"Photo #{existing_photo.id} '{picture_info.filename}' updates:"
                    )
                if attr == "user_id":
                    user_name = picture_info.user
                    logger.info(
                        f"  - {f'{attr}:':<15} from '{getattr(existing_photo, attr)}' to '{getattr(picture_info, attr)}' ({user_name})"
                    )
                else:
                    logger.info(
                        f"  - {f'{attr}:':<15} from '{getattr(existing_photo, attr)}' to '{getattr(picture_info, attr)}'"
                    )
                setattr(existing_photo, attr, getattr(picture_info, attr))
                something_changed = True
            elif getattr(picture_info, attr) is None:
                if something_changed is False:
                    logger.warning(
                        f"Photo #{existing_photo.id} '{picture_info.filename}' updates:"
                    )
                logger.warning(f"  - {f'{attr}:':<15} is NULL!")
                something_changed = True

        if something_changed:
            session.add(existing_photo)
            session.commit()
            return True, existing_photo.id
        else:
            return None, existing_photo.id

    # Insert new photo
    photo = Photo(
        filename=picture_info.filename,
        date_taken=picture_info.date_taken,
        album_id=album_id,
        user_id=picture_info.user_id,
        source_id=picture_info.source_id,
        saved_to_your_photos=picture_info.saved_to_your_photos,
    )
    session.add(photo)
    session.commit()
    logger.info(f"Added photo: {picture_info.filename}")
    return False, photo.id


def insert_error(
    session: Session, error_message: str, album_id: int | None = None
) -> None:
    """Insert an error."""
    error = Error(error_message=error_message, album_id=album_id)
    session.add(error)
    session.commit()
    logger.info(f"Error logged: {error_message}")


def link_user_to_album(session: Session, album_id: int, user_id: int) -> None:
    """Link a user to an album."""
    # Check if relationship already exists
    existing_link = (
        session.query(AlbumUser).filter_by(album_id=album_id, user_id=user_id).first()
    )

    if existing_link:
        logger.debug("User-album link already exists")
    else:
        link = AlbumUser(album_id=album_id, user_id=user_id)
        session.add(link)
        session.commit()
        logger.info("Linked user to album")


def album_exists(session: Session, album_title: str) -> bool:
    """Check if an album exists by title."""
    return (
        session.query(Album)
        .filter_by(source_title=album_title, source_type="gphoto")
        .first()
        is not None
    )


def get_album_photos_count(session: Session, album_id: int) -> int:
    """Get the number of photos for an album."""
    return session.query(Photo).filter_by(album_id=album_id).count()


def get_album_processed_items(session: Session, album_id: int) -> int:
    """Get the number of processed items for an album."""
    res = session.query(Album).filter_by(id=album_id).first()
    if res is not None:
        return res.processed_items
    return 0


def update_album_processed_items(
    session: Session, album_id: int, processed_items: int
) -> None:
    """Update the number of processed items for an album."""
    logger.debug(f"Updating album {album_id} processed items to {processed_items}")
    album = session.query(Album).filter_by(id=album_id).first()
    if album:
        album.processed_items = processed_items
        session.commit()


def is_album_fully_processed(session: Session, album_id: int) -> bool:
    """Check if an album is fully processed."""
    album = session.query(Album).filter_by(id=album_id).first()
    if album:
        return album.processed_items >= album.items
    return False


def get_albums_without_immich_id(session: Session) -> list[Album]:
    """Get all albums that don't have an immich_id set.

    Args:
        session: Database session

    Returns:
        List[Album]: List of Album objects without immich_id
    """
    return (
        session.query(Album)
        .filter(or_(Album.immich_id.is_(None), Album.immich_id == ""))
        .all()
    )


def get_photos_from_db(
    session: Session, album_id: int | None = None, has_immich_id: bool | None = None
) -> list[Photo]:
    """Get photos from the database with optional filtering.

    Args:
        session: Database session
        album_id: Optional album ID to filter photos by
        has_immich_id: If True, only return photos with immich_id set.
                       If False, only return photos without immich_id.
                       If None, return all photos (no filtering by immich_id).

    Returns:
        List[Photo]: List of Photo objects matching the criteria
    """
    query = session.query(Photo)

    if album_id is not None:
        query = query.filter(Photo.album_id == album_id)

    if has_immich_id is not None:
        if has_immich_id:
            query = query.filter(Photo.immich_id.is_not(None))
        else:
            query = query.filter(Photo.immich_id.is_(None))

    return query.all()


def get_users(session: Session) -> list[User]:
    """Get all users from database.

    Returns:
        List of User objects
    """
    return session.query(User).all()


def get_albums_from_db(
    session: Session,
    limit: int | None = None,
    offset: int = 0,
    not_finished: bool = False,
    album_ids: list[int] | None = None,
    not_saved: bool = False,
) -> list[AlbumInfo]:
    """Get albums from database with pagination.

    Args:
        session: Database session
        limit: Maximum number of albums to return
        offset: Number of albums to skip
        not_finished: If True, only return albums that are not fully processed
        not_saved: If True, only return albums that are not fully saved
    """
    # Subquery to count not saved photos per album
    not_saved_photos = (
        session.query(Photo.album_id, func.count(Photo.id).label("not_saved_count"))
        .filter(Photo.saved_to_your_photos == False)  # noqa: E712
        .group_by(Photo.album_id)
        .subquery()
    )

    query = session.query(
        Album.id,
        Album.source_url,
        Album.source_title,
        Album.items,
        Album.processed_items,
        Album.shared,
        Album.created_at,
        # Use case to determine if all photos are saved
        # If count of not_saved photos is None (no photos) or 0, then all are saved
        case(
            (func.coalesce(not_saved_photos.c.not_saved_count, 0) == 0, True),
            else_=False,
        ).label("all_photos_saved"),
    ).filter_by(source_type="gphoto")
    # Left join with the not_saved_photos subquery
    query = query.outerjoin(not_saved_photos, Album.id == not_saved_photos.c.album_id)
    if not_finished:
        query = query.filter(Album.processed_items < Album.items)
    if not_saved:
        query = query.filter(not_saved_photos.c.not_saved_count > 0)

    query = query.order_by(Album.id)

    if limit:
        query = query.offset(offset).limit(limit)

    if album_ids:
        query = query.filter(Album.id.in_(album_ids))

    res = query.all()
    return [
        AlbumInfo(
            album_id=album.id,
            title=album.source_title,
            items=album.items,
            shared=album.shared,
            processed_items=getattr(album, "processed_items", 0),
            created_at=getattr(album, "created_at", None),
            url=album.source_url,
            all_photos_saved=album.all_photos_saved,
        )
        for album in res
    ]


def get_users_from_db(session: Session) -> List[User]:
    """Get all users from database.

    Returns:
        List[User]: List of User objects with all fields including add_to_immich
    """
    return (
        session.query(User)
        .order_by(User.source_name)
        .filter_by(source_type="gphoto")
        .all()
    )


def get_database_stats(session: Session) -> Dict[str, Any]:
    """Get database statistics."""
    # Get album stats
    album_stats = (
        session.query(
            Album.id,
            Album.source_title,
            Album.source_type,
            Album.items,
            func.count(Photo.id).label("photo_count"),
            func.count(Error.id).label("error_count"),
        )
        .outerjoin(Photo, Album.id == Photo.album_id)
        .outerjoin(Error, Album.id == Error.album_id)
        .group_by(Album.id, Album.source_title, Album.source_type, Album.items)
        .all()
    )

    # Get user count
    user_count = session.query(User).count()

    # Get total photo count
    total_photos = session.query(Photo).count()

    # Get total error count
    total_errors = session.query(Error).count()

    return {
        "albums": album_stats,
        "total_albums": len(album_stats),
        "user_count": user_count,
        "total_users": user_count,
        "total_photos": total_photos,
        "total_errors": total_errors,
    }



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="DownloadInfoDto")


@_attrs_define
class DownloadInfoDto:
    """
    Attributes:
        album_id (Union[Unset, UUID]):
        archive_size (Union[Unset, int]):
        asset_ids (Union[Unset, list[UUID]]):
        user_id (Union[Unset, UUID]):
    """

    album_id: Union[Unset, UUID] = UNSET
    archive_size: Union[Unset, int] = UNSET
    asset_ids: Union[Unset, list[UUID]] = UNSET
    user_id: Union[Unset, UUID] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        album_id: Union[Unset, str] = UNSET
        if not isinstance(self.album_id, Unset):
            album_id = str(self.album_id)

        archive_size = self.archive_size

        asset_ids: Union[Unset, list[str]] = UNSET
        if not isinstance(self.asset_ids, Unset):
            asset_ids = []
            for asset_ids_item_data in self.asset_ids:
                asset_ids_item = str(asset_ids_item_data)
                asset_ids.append(asset_ids_item)

        user_id: Union[Unset, str] = UNSET
        if not isinstance(self.user_id, Unset):
            user_id = str(self.user_id)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if album_id is not UNSET:
            field_dict["albumId"] = album_id
        if archive_size is not UNSET:
            field_dict["archiveSize"] = archive_size
        if asset_ids is not UNSET:
            field_dict["assetIds"] = asset_ids
        if user_id is not UNSET:
            field_dict["userId"] = user_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        _album_id = d.pop("albumId", UNSET)
        album_id: Union[Unset, UUID]
        if isinstance(_album_id, Unset):
            album_id = UNSET
        else:
            album_id = UUID(_album_id)

        archive_size = d.pop("archiveSize", UNSET)

        asset_ids = []
        _asset_ids = d.pop("assetIds", UNSET)
        for asset_ids_item_data in _asset_ids or []:
            asset_ids_item = UUID(asset_ids_item_data)

            asset_ids.append(asset_ids_item)

        _user_id = d.pop("userId", UNSET)
        user_id: Union[Unset, UUID]
        if isinstance(_user_id, Unset):
            user_id = UNSET
        else:
            user_id = UUID(_user_id)

        download_info_dto = cls(
            album_id=album_id,
            archive_size=archive_size,
            asset_ids=asset_ids,
            user_id=user_id,
        )

        download_info_dto.additional_properties = d
        return download_info_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.shared_link_type import SharedLinkType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.album_response_dto import AlbumResponseDto
    from ..models.asset_response_dto import AssetResponseDto


T = TypeVar("T", bound="SharedLinkResponseDto")


@_attrs_define
class SharedLinkResponseDto:
    """
    Attributes:
        allow_download (bool):
        allow_upload (bool):
        assets (list['AssetResponseDto']):
        created_at (datetime.datetime):
        description (Union[None, str]):
        expires_at (Union[None, datetime.datetime]):
        id (str):
        key (str):
        password (Union[None, str]):
        show_metadata (bool):
        slug (Union[None, str]):
        type_ (SharedLinkType):
        user_id (str):
        album (Union[Unset, AlbumResponseDto]):
        token (Union[None, Unset, str]):
    """

    allow_download: bool
    allow_upload: bool
    assets: list["AssetResponseDto"]
    created_at: datetime.datetime
    description: Union[None, str]
    expires_at: Union[None, datetime.datetime]
    id: str
    key: str
    password: Union[None, str]
    show_metadata: bool
    slug: Union[None, str]
    type_: SharedLinkType
    user_id: str
    album: Union[Unset, "AlbumResponseDto"] = UNSET
    token: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        allow_download = self.allow_download

        allow_upload = self.allow_upload

        assets = []
        for assets_item_data in self.assets:
            assets_item = assets_item_data.to_dict()
            assets.append(assets_item)

        created_at = self.created_at.isoformat()

        description: Union[None, str]
        description = self.description

        expires_at: Union[None, str]
        if isinstance(self.expires_at, datetime.datetime):
            expires_at = self.expires_at.isoformat()
        else:
            expires_at = self.expires_at

        id = self.id

        key = self.key

        password: Union[None, str]
        password = self.password

        show_metadata = self.show_metadata

        slug: Union[None, str]
        slug = self.slug

        type_ = self.type_.value

        user_id = self.user_id

        album: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.album, Unset):
            album = self.album.to_dict()

        token: Union[None, Unset, str]
        if isinstance(self.token, Unset):
            token = UNSET
        else:
            token = self.token

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "allowDownload": allow_download,
                "allowUpload": allow_upload,
                "assets": assets,
                "createdAt": created_at,
                "description": description,
                "expiresAt": expires_at,
                "id": id,
                "key": key,
                "password": password,
                "showMetadata": show_metadata,
                "slug": slug,
                "type": type_,
                "userId": user_id,
            }
        )
        if album is not UNSET:
            field_dict["album"] = album
        if token is not UNSET:
            field_dict["token"] = token

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.album_response_dto import AlbumResponseDto
        from ..models.asset_response_dto import AssetResponseDto

        d = dict(src_dict)
        allow_download = d.pop("allowDownload")

        allow_upload = d.pop("allowUpload")

        assets = []
        _assets = d.pop("assets")
        for assets_item_data in _assets:
            assets_item = AssetResponseDto.from_dict(assets_item_data)

            assets.append(assets_item)

        created_at = isoparse(d.pop("createdAt"))

        def _parse_description(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        description = _parse_description(d.pop("description"))

        def _parse_expires_at(data: object) -> Union[None, datetime.datetime]:
            if data is None:
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                expires_at_type_0 = isoparse(data)

                return expires_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, datetime.datetime], data)

        expires_at = _parse_expires_at(d.pop("expiresAt"))

        id = d.pop("id")

        key = d.pop("key")

        def _parse_password(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        password = _parse_password(d.pop("password"))

        show_metadata = d.pop("showMetadata")

        def _parse_slug(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        slug = _parse_slug(d.pop("slug"))

        type_ = SharedLinkType(d.pop("type"))

        user_id = d.pop("userId")

        _album = d.pop("album", UNSET)
        album: Union[Unset, AlbumResponseDto]
        if isinstance(_album, Unset):
            album = UNSET
        else:
            album = AlbumResponseDto.from_dict(_album)

        def _parse_token(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        token = _parse_token(d.pop("token", UNSET))

        shared_link_response_dto = cls(
            allow_download=allow_download,
            allow_upload=allow_upload,
            assets=assets,
            created_at=created_at,
            description=description,
            expires_at=expires_at,
            id=id,
            key=key,
            password=password,
            show_metadata=show_metadata,
            slug=slug,
            type_=type_,
            user_id=user_id,
            album=album,
            token=token,
        )

        shared_link_response_dto.additional_properties = d
        return shared_link_response_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ManualJobName(str, Enum):
    BACKUP_DATABASE = "backup-database"
    MEMORY_CLEANUP = "memory-cleanup"
    MEMORY_CREATE = "memory-create"
    PERSON_CLEANUP = "person-cleanup"
    TAG_CLEANUP = "tag-cleanup"
    USER_CLEANUP = "user-cleanup"

    def __str__(self) -> str:
        return str(self.value)



from collections.abc import Mapping
from typing import Any, TypeVar, TYPE_CHECKING

from attrs import define as _attrs_define
from attrs import field as _attrs_field
import json
from .. import types

from ..types import UNSET, Unset

from ..models.asset_visibility import AssetVisibility
from ..types import File, FileTypes
from dateutil.parser import isoparse
from io import BytesIO
from typing import Union
from uuid import UUID
import datetime

if TYPE_CHECKING:
    from ..models.asset_metadata_upsert_item_dto import AssetMetadataUpsertItemDto


T = TypeVar("T", bound="AssetMediaCreateDto")


@_attrs_define
class AssetMediaCreateDto:
    """
    Attributes:
        asset_data (File):
        device_asset_id (str):
        device_id (str):
        file_created_at (datetime.datetime):
        file_modified_at (datetime.datetime):
        metadata (list['AssetMetadataUpsertItemDto']):
        duration (Union[Unset, str]):
        filename (Union[Unset, str]):
        is_favorite (Union[Unset, bool]):
        live_photo_video_id (Union[Unset, UUID]):
        sidecar_data (Union[Unset, File]):
        visibility (Union[Unset, AssetVisibility]):
    """

    asset_data: File
    device_asset_id: str
    device_id: str
    file_created_at: datetime.datetime
    file_modified_at: datetime.datetime
    metadata: list["AssetMetadataUpsertItemDto"]
    duration: Union[Unset, str] = UNSET
    filename: Union[Unset, str] = UNSET
    is_favorite: Union[Unset, bool] = UNSET
    live_photo_video_id: Union[Unset, UUID] = UNSET
    sidecar_data: Union[Unset, File] = UNSET
    visibility: Union[Unset, AssetVisibility] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        asset_data = self.asset_data.to_tuple()

        device_asset_id = self.device_asset_id

        device_id = self.device_id

        file_created_at = self.file_created_at.isoformat()

        file_modified_at = self.file_modified_at.isoformat()

        metadata = []
        for metadata_item_data in self.metadata:
            metadata_item = metadata_item_data.to_dict()
            metadata.append(metadata_item)

        duration = self.duration

        filename = self.filename

        is_favorite = self.is_favorite

        live_photo_video_id: Union[Unset, str] = UNSET
        if not isinstance(self.live_photo_video_id, Unset):
            live_photo_video_id = str(self.live_photo_video_id)

        sidecar_data: Union[Unset, FileTypes] = UNSET
        if not isinstance(self.sidecar_data, Unset):
            sidecar_data = self.sidecar_data.to_tuple()

        visibility: Union[Unset, str] = UNSET
        if not isinstance(self.visibility, Unset):
            visibility = self.visibility.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "assetData": asset_data,
                "deviceAssetId": device_asset_id,
                "deviceId": device_id,
                "fileCreatedAt": file_created_at,
                "fileModifiedAt": file_modified_at,
                "metadata": metadata,
            }
        )
        if duration is not UNSET:
            field_dict["duration"] = duration
        if filename is not UNSET:
            field_dict["filename"] = filename
        if is_favorite is not UNSET:
            field_dict["isFavorite"] = is_favorite
        if live_photo_video_id is not UNSET:
            field_dict["livePhotoVideoId"] = live_photo_video_id
        if sidecar_data is not UNSET:
            field_dict["sidecarData"] = sidecar_data
        if visibility is not UNSET:
            field_dict["visibility"] = visibility

        return field_dict

    def to_multipart(self) -> types.RequestFiles:
        files: types.RequestFiles = []

        files.append(("assetData", self.asset_data.to_tuple()))

        files.append(
            ("deviceAssetId", (None, str(self.device_asset_id).encode(), "text/plain"))
        )

        files.append(("deviceId", (None, str(self.device_id).encode(), "text/plain")))

        files.append(
            (
                "fileCreatedAt",
                (None, self.file_created_at.isoformat().encode(), "text/plain"),
            )
        )

        files.append(
            (
                "fileModifiedAt",
                (None, self.file_modified_at.isoformat().encode(), "text/plain"),
            )
        )

        for metadata_item_element in self.metadata:
            files.append(
                (
                    "metadata",
                    (
                        None,
                        json.dumps(metadata_item_element.to_dict()).encode(),
                        "application/json",
                    ),
                )
            )

        if not isinstance(self.duration, Unset):
            files.append(
                ("duration", (None, str(self.duration).encode(), "text/plain"))
            )

        if not isinstance(self.filename, Unset):
            files.append(
                ("filename", (None, str(self.filename).encode(), "text/plain"))
            )

        if not isinstance(self.is_favorite, Unset):
            files.append(
                ("isFavorite", (None, str(self.is_favorite).encode(), "text/plain"))
            )

        if not isinstance(self.live_photo_video_id, Unset):
            files.append(
                (
                    "livePhotoVideoId",
                    (None, str(self.live_photo_video_id), "text/plain"),
                )
            )

        if not isinstance(self.sidecar_data, Unset):
            files.append(("sidecarData", self.sidecar_data.to_tuple()))

        if not isinstance(self.visibility, Unset):
            files.append(
                (
                    "visibility",
                    (None, str(self.visibility.value).encode(), "text/plain"),
                )
            )

        for prop_name, prop in self.additional_properties.items():
            files.append((prop_name, (None, str(prop).encode(), "text/plain")))

        return files

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.asset_metadata_upsert_item_dto import AssetMetadataUpsertItemDto

        d = dict(src_dict)
        asset_data = File(payload=BytesIO(d.pop("assetData")))

        device_asset_id = d.pop("deviceAssetId")

        device_id = d.pop("deviceId")

        file_created_at = isoparse(d.pop("fileCreatedAt"))

        file_modified_at = isoparse(d.pop("fileModifiedAt"))

        metadata = []
        _metadata = d.pop("metadata")
        for metadata_item_data in _metadata:
            metadata_item = AssetMetadataUpsertItemDto.from_dict(metadata_item_data)

            metadata.append(metadata_item)

        duration = d.pop("duration", UNSET)

        filename = d.pop("filename", UNSET)

        is_favorite = d.pop("isFavorite", UNSET)

        _live_photo_video_id = d.pop("livePhotoVideoId", UNSET)
        live_photo_video_id: Union[Unset, UUID]
        if isinstance(_live_photo_video_id, Unset):
            live_photo_video_id = UNSET
        else:
            live_photo_video_id = UUID(_live_photo_video_id)

        _sidecar_data = d.pop("sidecarData", UNSET)
        sidecar_data: Union[Unset, File]
        if isinstance(_sidecar_data, Unset):
            sidecar_data = UNSET
        else:
            sidecar_data = File(payload=BytesIO(_sidecar_data))

        _visibility = d.pop("visibility", UNSET)
        visibility: Union[Unset, AssetVisibility]
        if isinstance(_visibility, Unset):
            visibility = UNSET
        else:
            visibility = AssetVisibility(_visibility)

        asset_media_create_dto = cls(
            asset_data=asset_data,
            device_asset_id=device_asset_id,
            device_id=device_id,
            file_created_at=file_created_at,
            file_modified_at=file_modified_at,
            metadata=metadata,
            duration=duration,
            filename=filename,
            is_favorite=is_favorite,
            live_photo_video_id=live_photo_video_id,
            sidecar_data=sidecar_data,
            visibility=visibility,
        )

        asset_media_create_dto.additional_properties = d
        return asset_media_create_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from immichporter.immich.client.models.bulk_ids_dto import BulkIdsDto
from immichporter.immich.client.models.asset_response_dto import AssetResponseDto
from immichporter.immich.client import AuthenticatedClient
from typing import Type, Any
from immichporter.immich.client.api.albums import (
    get_all_albums,
    create_album,
    delete_album,
)
import time
import re
from immichporter.immich.client.api.timeline import get_time_bucket, get_time_buckets
from immichporter.immich.client.api.tags import (
    get_all_tags,
    untag_assets,
    delete_tag,
    get_tag_by_id,
)
from immichporter.immich.client.api.search import search_assets
from immichporter.immich.client.api.users_admin import (
    search_users_admin,
    create_user_admin,
)
from uuid import UUID
from immichporter.immich.client.models import (
    TagResponseDto,
    MetadataSearchDto,
    UserAdminCreateDto,
    AlbumUserCreateDto,
    CreateAlbumDto,
    AlbumResponseDto,
    JobCreateDto,
    UserAdminResponseDto,
    JobCommandDto,
    ManualJobName,
    JobName,
    JobCommand,
    JobStatusDto,
)
from immichporter.immich.client.api.jobs import (
    create_job,
    get_all_jobs_status,
    send_job_command,
)
from immichporter.immich.client.types import UNSET, Unset
from rich.console import Console
from datetime import datetime, timedelta
from loguru import logger

console = Console()
ImmichApiClient: Type[AuthenticatedClient] = AuthenticatedClient


def immich_api_client(
    endpoint: str, api_key: str, insecure: bool = False
) -> ImmichApiClient:
    """Returns immich api client"""
    base_url = endpoint.rstrip("/")
    if not base_url.endswith("/api"):
        base_url = f"{base_url}/api"
    client = AuthenticatedClient(
        base_url=base_url,  # type: ignore
        token=api_key,
        auth_header_name="x-api-key",
        prefix="",
        verify_ssl=not insecure,  # type: ignore
    )

    return client


class ImmichClient:
    def __init__(
        self,
        client: ImmichApiClient | None = None,
        endpoint: str | None = None,
        api_key: str | None = None,
        insecure: bool = False,
    ):
        """Immich client with specific functions, often an API wrapper."""
        self._api_key = api_key
        if client is not None:
            self._client = client
        else:
            assert (
                endpoint is not None
            ), "'endpoint' must be provided if 'client' is not provided"
            assert (
                api_key is not None
            ), "'api_key' must be provided if 'client' is not provided"
            self._client = immich_api_client(
                endpoint=endpoint, api_key=api_key, insecure=insecure
            )

    @property
    def client(self) -> ImmichApiClient:
        return self._client

    @property
    def endpoint(self) -> str:
        """Returns the base url of the Immich server"""
        return self.client._base_url

    def get_albums(
        self, limit: int | None = None, shared: bool | None = None
    ) -> list[AlbumResponseDto]:
        """List all albums on the Immich server.

        Args:
            limit: Maximum number of albums to return
            shared: Filter by shared status (True for shared, False for not shared, None for all)
        """
        shared: Unset | bool = UNSET if shared is None else shared
        response = get_all_albums.sync_detailed(client=self.client, shared=shared)
        if response.status_code != 200:
            raise Exception(f"Failed to fetch albums: {response.content}")

        assert response.parsed is not None
        albums: list[AlbumResponseDto] = response.parsed

        # Sort albums by name
        albums = sorted(albums, key=lambda x: x.album_name.lower())

        # Apply limit
        if limit is not None and limit > 0:
            albums = albums[:limit]

        return albums

    def create_album(
        self,
        name: str,
        description: str | None = None,
        users: list[AlbumUserCreateDto] | None = None,
        assets: list[str] | None = None,
    ) -> AlbumResponseDto:
        description: Unset | str = UNSET if description is None else description
        users: Unset | list[AlbumUserCreateDto] = UNSET if users is None else users
        assets: Unset | list[UUID] = (
            UNSET if assets is None else [UUID(a) for a in assets]
        )
        body = CreateAlbumDto(
            album_name=name,
            description=description,
            album_users=users,
            asset_ids=assets,
        )
        response = create_album.sync_detailed(client=self.client, body=body)
        if response.status_code != 201:
            raise Exception(
                f"Failed to create album {response.status_code}: {response.content}"
            )
        assert response.parsed is not None
        return response.parsed

    def delete_album(
        self, album_id: str | None = None, album_name: str | None = None
    ) -> None:
        if album_id is None and album_name is None:
            raise ValueError("Either album_id or album_name must be provided")
        if album_id is None:
            albums = self.get_albums()
            for album in albums:
                if album.album_name == album_name:
                    album_id = album.id
                    break
        if album_id is None:
            raise ValueError(f"Album '{album_name}' not found")
        response = delete_album.sync_detailed(client=self.client, id=UUID(album_id))
        if response.status_code != 204:
            raise Exception(
                f"Failed to delete album {response.status_code}: {response.content}"
            )

    def search_assets(
        self,
        filename: str | None | Unset = None,
        taken: datetime | str | None | Unset = None,
        taken_before: datetime | str | None | Unset = None,
        taken_after: datetime | str | None | Unset = None,
        **options: dict[str, Any],
    ) -> list[AssetResponseDto]:
        """Search for assets on the Immich server.

        Dates can be formate as follow:
        Python `datetime` or string with the format `%Y-%m-%d %H:%M:%S` or `%Y-%m-%d`

        Args:
            filename: Filter by filename
            taken: Filter by taken date (plus minus 1 day if no time is given, resp. minus 2 hours if no day is given)
            taken_before: Filter by taken date before, cannot be used together with `taken`.
            taken_after: Filter by taken date after, cannot be used together with `taken`.
            **options: Additional options, see https://api.immich.app/endpoints/search/searchAssets for more information
        """
        filename = UNSET if filename is None else filename
        taken_before = UNSET if taken_before is None else taken_before
        taken_after = UNSET if taken_after is None else taken_after
        if isinstance(taken_before, str):
            if " " not in taken_before:
                taken_before += " 00:00:00"
            taken_before = datetime.strptime(taken_before, "%Y-%m-%d %H:%M:%S")
        if isinstance(taken_after, str):
            if " " not in taken_after:
                taken_after += " 00:00:00"
            taken_after = datetime.strptime(taken_after, "%Y-%m-%d %H:%M:%S")
        if taken:
            assert (
                taken_before is UNSET and taken_after is UNSET
            ), "'taken_before' and 'taken_after' must be unset if 'taken' is set"
            delta_before = timedelta(hours=2)
            delta_after = timedelta(hours=2)
            if isinstance(taken, str):
                if " " not in taken:
                    taken += " 00:00:00"
                    delta_before = timedelta(days=1)
                    delta_after = timedelta(days=0)
                taken = datetime.strptime(taken, "%Y-%m-%d %H:%M:%S")
            taken_before = taken + delta_before
            taken_after = taken - delta_after

        search_dto = MetadataSearchDto(
            original_file_name=filename,
            taken_before=taken_before,
            taken_after=taken_after,
            **options,  # type: ignore
        )
        response = search_assets.sync_detailed(client=self.client, body=search_dto)
        if response.status_code != 200:
            raise Exception(f"Failed to fetch albums: {response.content}")

        assets = response.parsed
        assert assets is not None

        return assets.assets.items

    def get_users(self, width_deleted: bool = True) -> list[UserAdminResponseDto]:
        response = search_users_admin.sync_detailed(client=self.client)
        if response.status_code != 200:
            raise Exception(f"Failed to fetch users: {response.content}")
        assert response.parsed is not None
        return response.parsed

    def add_user(
        self, name: str, email: str, password: str, quota_gb: int = 10
    ) -> UserAdminResponseDto:
        quota_bytes = 1073741824 * quota_gb
        body = UserAdminCreateDto(
            name=name,
            email=email,
            password=password,
            notify=False,
            should_change_password=True,
            quota_size_in_bytes=quota_bytes,
        )
        response = create_user_admin.sync_detailed(client=self.client, body=body)
        if response.status_code != 201:
            raise Exception(
                f"Failed to create user {response.status_code}: {response.content}"
            )
        res = response.parsed
        assert res is not None
        return res

    def start_job(self, job_name: ManualJobName | JobName) -> bool:
        """Start a new job with the given name.

        Args:
            job_name: Name of the job to start (must be a valid ManualJobName or JobName value)

        Returns:
            dict: The job status with keys: id, name, isActive, lastRun, nextRun

        Raises:
            Exception: If the job fails to start or status cannot be retrieved
        """
        try:
            # Create the job data
            if isinstance(job_name, ManualJobName):
                job_data = JobCreateDto(name=ManualJobName(job_name))

                # Send the request to create the job
                response = create_job.sync_detailed(client=self._client, body=job_data)

                # 204 No Content is expected on success
                if response.status_code != 204:
                    raise Exception(
                        f"Failed to start job {job_name}: {response.content}"
                    )
                return True
            elif isinstance(job_name, JobName):
                # Send the request to create the job
                body = JobCommandDto(command=JobCommand.START)
                response = send_job_command.sync_detailed(
                    id=job_name, client=self._client, body=body
                )
                if response.status_code == 400:
                    if (
                        "job is already running"
                        in response.content.decode("utf-8").lower()
                    ):
                        console.print(
                            f"Job [blue]'{job_name}'[/] is already running, maybe you need to restart it later manually!"
                        )
                        return True

                if not response.status_code.is_success:  # type: ignore
                    raise Exception(
                        f"Failed to start job {job_name} (status code: {response.status_code}): {response.content}"
                    )
                job_status = response.parsed
                assert job_status is not None
                return job_status.job_counts.active >= 0

        except Exception as e:
            raise Exception(f"Error starting job {job_name}: {str(e)}") from e

    def get_job_status(self, job_name: JobName) -> JobStatusDto:
        """Get the status of a job by name.

        Args:
            job_name: Name of the job to check

        Returns:
            dict: The job status with keys: id, name, isActive, lastRun, nextRun

        Raises:
            Exception: If the job status cannot be retrieved
        """
        try:
            response = get_all_jobs_status.sync_detailed(client=self._client)
            if not response.status_code.is_success:  # type: ignore
                raise Exception(
                    f"Failed to get job status (status code: {response.status_code}): {response.content}"
                )

            resp_parsed = response.parsed
            job_name_snake = re.sub(r"([a-z])([A-Z])", r"\1_\2", job_name).lower()
            job_status = getattr(resp_parsed, job_name_snake, None)
            if not job_status:
                raise Exception(f"Could not find status for job {job_name}")

            return job_status

        except Exception as e:
            raise Exception(f"Error getting job status for {job_name}: {str(e)}")

    def run_db_backup(self, wait_time_s: int = 15):
        self.start_job(ManualJobName.BACKUP_DATABASE)
        time.sleep(wait_time_s)

    def get_tags(
        self,
        filter_name: str | None = None,
        filter_value: str | None = None,
        parent_id: UUID | str | None = None,
    ) -> list[TagResponseDto]:
        """Get all tags from Immich.

        Args:
            filter_name (str | None, optional): Filter tags by name (e.g. `TagName`). Defaults to None.
            filter_value (str | None, optional): Filter tags by value (parents included, e.g. `Parent/TagName`). Defaults to None.
            parent_id (UUID | str | None, optional): Filter tags by parent ID (either UUID or name, e.g. `Parent`). Defaults to None.

        Returns:
            List of tags
        """
        response = get_all_tags.sync_detailed(client=self._client)
        if not response.status_code.is_success:  # type: ignore
            raise Exception(
                f"Failed to get tags (status code: {response.status_code}): {response.content}"
            )
        tags = response.parsed
        if not tags:
            raise Exception("Could not find tags")
        if parent_id:
            if isinstance(parent_id, UUID):
                parent_uuid = str(parent_id)
            else:
                try:
                    parent_uuid = UUID(parent_id)
                except ValueError:
                    parent_tags = self.get_tags(filter=parent_id)
                    if not parent_tags:
                        raise Exception(f"Could not find tag with name {parent_id}")
                    parent_uuid = parent_tags[0].id
            tags = [tag for tag in tags if tag.parent_id == str(parent_uuid)]
        if filter_value:
            tags = [tag for tag in tags if filter_value in tag.value]
        if filter_name:
            tags = [tag for tag in tags if filter_name in tag.name]
        return tags

    def timeline_assets(self, tag_id: UUID | str | None = None, **kwargs):
        """Get all assets from Immich by tag ID or **kwargs."""
        tag_id_arg = (
            tag_id if isinstance(tag_id, UUID) else UUID(tag_id) if tag_id else UNSET
        )
        response = get_time_buckets.sync_detailed(
            client=self._client, tag_id=tag_id_arg, **kwargs
        )
        if not response.status_code.is_success:  # type: ignore
            raise Exception(
                f"Failed to get time buckets (status code: {response.status_code}): {response.content}"
            )
        time_buckets = response.parsed
        if time_buckets is None:
            raise Exception("Could not find time buckets")
        for time_bucket in time_buckets:
            time_bucket_ts = time_bucket.time_bucket
            resp = get_time_bucket.sync_detailed(
                client=self._client,
                tag_id=tag_id_arg,
                time_bucket=time_bucket_ts,
                **kwargs,
            )
            if not resp.status_code.is_success:  # type: ignore
                raise Exception(
                    f"Failed to get time bucket (status code: {resp.status_code}): {resp.content}"
                )
            resp_parsed = resp.parsed
            if not resp_parsed:
                raise Exception("Could not find time bucket")
            for asset_id in resp_parsed.id:
                yield asset_id

    def untag_assets(
        self,
        tag: TagResponseDto | str | UUID,
        asset_ids: list[UUID | str] | None = None,
        remove_tag: bool = False,
    ) -> int:
        if isinstance(tag, str):
            tag = self.get_tags(filter_name=tag)[0].id
        elif isinstance(tag, UUID):
            res = get_tag_by_id.sync(client=self._client, id=tag)
            if res is None:
                raise Exception(f"Could not find tag with id {tag}")
            tag = res
        if isinstance(tag, TagResponseDto):
            tag_id = tag.id
        else:
            raise Exception(f"Invalid tag type: {type(tag)}")

        removed = 0
        # check if there are children:
        children = self.get_tags(parent_id=tag_id)
        if children:
            logger.info(f"Tag {tag_id} has children, removing them first")
            # we need to remove children first:
            for child in children:
                removed += self.untag_assets(tag=child, remove_tag=remove_tag)

        if asset_ids is None:
            asset_ids = list(self.timeline_assets(tag_id=tag_id))
        asset_ids = [
            UUID(asset_id) if isinstance(asset_id, str) else asset_id
            for asset_id in asset_ids
        ]
        logger.debug(
            f"Untagging {len(asset_ids)} assets with tag '{tag.value}' ({tag_id})"
        )
        tag_id = UUID(tag_id) if isinstance(tag_id, str) else tag_id
        body = BulkIdsDto(ids=asset_ids)
        response = untag_assets.sync_detailed(client=self._client, id=tag_id, body=body)
        if not response.status_code.is_success:  # type: ignore
            raise Exception(
                f"Failed to untag assets (status code: {response.status_code}): {response.content}"
            )
        removed += len(asset_ids)
        if remove_tag:
            logger.debug(f"Removing tag '{tag.value}' ({tag_id})")
            response = delete_tag.sync_detailed(client=self._client, id=tag_id)
            if not response.status_code.is_success:  # type: ignore
                raise Exception(
                    f"Failed to delete tag (status code: {response.status_code}): {response.content}"
                )
        return removed


if __name__ == "__main__":
    import os

    endpoint = os.getenv("IMMICH_ENDPOINT")
    api_key = os.getenv("IMMICH_API_KEY")
    insecure = os.getenv("IMMICH_INSECURE") == "1"
    client = ImmichClient(endpoint=endpoint, api_key=api_key, insecure=insecure)
    console.print(f"Endpoint: {client.endpoint}")
    console.print(f"API Key: [yellow]'{client._api_key}'[/]")
    tag_filter = "from_google"
    tag_filter = None
    # tag_parent_id = UUID("a94b0135-727d-4f32-9b43-c99d5ac7d92e")
    # tag_parent_id = "People"
    tag_parent_id = None
    tags = client.get_tags(filter_name=tag_filter, parent_id=tag_parent_id)
    for tag in tags:
        console.print(tag.value)
    # tag = tags[0]
    ##console.print(tags)
    # console.print(tag)
    # removed = client.untag_assets(tag=tag, remove_tag=True)
    # console.print(f"Removed {removed} assets")
    ##for asset_id in client.search_assets(tag_id=tag.id):
    #    #console.print(asset_id)



from http import HTTPStatus
from typing import Any, Optional, Union
from uuid import UUID

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.people_response_dto import PeopleResponseDto
from ...types import UNSET, Response, Unset


def _get_kwargs(
    *,
    closest_asset_id: Union[Unset, UUID] = UNSET,
    closest_person_id: Union[Unset, UUID] = UNSET,
    page: Union[Unset, float] = 1.0,
    size: Union[Unset, float] = 500.0,
    with_hidden: Union[Unset, bool] = UNSET,
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    json_closest_asset_id: Union[Unset, str] = UNSET
    if not isinstance(closest_asset_id, Unset):
        json_closest_asset_id = str(closest_asset_id)
    params["closestAssetId"] = json_closest_asset_id

    json_closest_person_id: Union[Unset, str] = UNSET
    if not isinstance(closest_person_id, Unset):
        json_closest_person_id = str(closest_person_id)
    params["closestPersonId"] = json_closest_person_id

    params["page"] = page

    params["size"] = size

    params["withHidden"] = with_hidden

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": "/people",
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[PeopleResponseDto]:
    if response.status_code == 200:
        response_200 = PeopleResponseDto.from_dict(response.json())

        return response_200

    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[PeopleResponseDto]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
    closest_asset_id: Union[Unset, UUID] = UNSET,
    closest_person_id: Union[Unset, UUID] = UNSET,
    page: Union[Unset, float] = 1.0,
    size: Union[Unset, float] = 500.0,
    with_hidden: Union[Unset, bool] = UNSET,
) -> Response[PeopleResponseDto]:
    """This endpoint requires the `person.read` permission.

    Args:
        closest_asset_id (Union[Unset, UUID]):
        closest_person_id (Union[Unset, UUID]):
        page (Union[Unset, float]):  Default: 1.0.
        size (Union[Unset, float]):  Default: 500.0.
        with_hidden (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[PeopleResponseDto]
    """

    kwargs = _get_kwargs(
        closest_asset_id=closest_asset_id,
        closest_person_id=closest_person_id,
        page=page,
        size=size,
        with_hidden=with_hidden,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient,
    closest_asset_id: Union[Unset, UUID] = UNSET,
    closest_person_id: Union[Unset, UUID] = UNSET,
    page: Union[Unset, float] = 1.0,
    size: Union[Unset, float] = 500.0,
    with_hidden: Union[Unset, bool] = UNSET,
) -> Optional[PeopleResponseDto]:
    """This endpoint requires the `person.read` permission.

    Args:
        closest_asset_id (Union[Unset, UUID]):
        closest_person_id (Union[Unset, UUID]):
        page (Union[Unset, float]):  Default: 1.0.
        size (Union[Unset, float]):  Default: 500.0.
        with_hidden (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        PeopleResponseDto
    """

    return sync_detailed(
        client=client,
        closest_asset_id=closest_asset_id,
        closest_person_id=closest_person_id,
        page=page,
        size=size,
        with_hidden=with_hidden,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
    closest_asset_id: Union[Unset, UUID] = UNSET,
    closest_person_id: Union[Unset, UUID] = UNSET,
    page: Union[Unset, float] = 1.0,
    size: Union[Unset, float] = 500.0,
    with_hidden: Union[Unset, bool] = UNSET,
) -> Response[PeopleResponseDto]:
    """This endpoint requires the `person.read` permission.

    Args:
        closest_asset_id (Union[Unset, UUID]):
        closest_person_id (Union[Unset, UUID]):
        page (Union[Unset, float]):  Default: 1.0.
        size (Union[Unset, float]):  Default: 500.0.
        with_hidden (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[PeopleResponseDto]
    """

    kwargs = _get_kwargs(
        closest_asset_id=closest_asset_id,
        closest_person_id=closest_person_id,
        page=page,
        size=size,
        with_hidden=with_hidden,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient,
    closest_asset_id: Union[Unset, UUID] = UNSET,
    closest_person_id: Union[Unset, UUID] = UNSET,
    page: Union[Unset, float] = 1.0,
    size: Union[Unset, float] = 500.0,
    with_hidden: Union[Unset, bool] = UNSET,
) -> Optional[PeopleResponseDto]:
    """This endpoint requires the `person.read` permission.

    Args:
        closest_asset_id (Union[Unset, UUID]):
        closest_person_id (Union[Unset, UUID]):
        page (Union[Unset, float]):  Default: 1.0.
        size (Union[Unset, float]):  Default: 500.0.
        with_hidden (Union[Unset, bool]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        PeopleResponseDto
    """

    return (
        await asyncio_detailed(
            client=client,
            closest_asset_id=closest_asset_id,
            closest_person_id=closest_person_id,
            page=page,
            size=size,
            with_hidden=with_hidden,
        )
    ).parsed



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="LoginResponseDto")


@_attrs_define
class LoginResponseDto:
    """
    Attributes:
        access_token (str):
        is_admin (bool):
        is_onboarded (bool):
        name (str):
        profile_image_path (str):
        should_change_password (bool):
        user_email (str):
        user_id (str):
    """

    access_token: str
    is_admin: bool
    is_onboarded: bool
    name: str
    profile_image_path: str
    should_change_password: bool
    user_email: str
    user_id: str
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        access_token = self.access_token

        is_admin = self.is_admin

        is_onboarded = self.is_onboarded

        name = self.name

        profile_image_path = self.profile_image_path

        should_change_password = self.should_change_password

        user_email = self.user_email

        user_id = self.user_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "accessToken": access_token,
                "isAdmin": is_admin,
                "isOnboarded": is_onboarded,
                "name": name,
                "profileImagePath": profile_image_path,
                "shouldChangePassword": should_change_password,
                "userEmail": user_email,
                "userId": user_id,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        access_token = d.pop("accessToken")

        is_admin = d.pop("isAdmin")

        is_onboarded = d.pop("isOnboarded")

        name = d.pop("name")

        profile_image_path = d.pop("profileImagePath")

        should_change_password = d.pop("shouldChangePassword")

        user_email = d.pop("userEmail")

        user_id = d.pop("userId")

        login_response_dto = cls(
            access_token=access_token,
            is_admin=is_admin,
            is_onboarded=is_onboarded,
            name=name,
            profile_image_path=profile_image_path,
            should_change_password=should_change_password,
            user_email=user_email,
            user_id=user_id,
        )

        login_response_dto.additional_properties = d
        return login_response_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

T = TypeVar("T", bound="TagResponseDto")


@_attrs_define
class TagResponseDto:
    """
    Attributes:
        created_at (datetime.datetime):
        id (str):
        name (str):
        updated_at (datetime.datetime):
        value (str):
        color (Union[Unset, str]):
        parent_id (Union[Unset, str]):
    """

    created_at: datetime.datetime
    id: str
    name: str
    updated_at: datetime.datetime
    value: str
    color: Union[Unset, str] = UNSET
    parent_id: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        created_at = self.created_at.isoformat()

        id = self.id

        name = self.name

        updated_at = self.updated_at.isoformat()

        value = self.value

        color = self.color

        parent_id = self.parent_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "createdAt": created_at,
                "id": id,
                "name": name,
                "updatedAt": updated_at,
                "value": value,
            }
        )
        if color is not UNSET:
            field_dict["color"] = color
        if parent_id is not UNSET:
            field_dict["parentId"] = parent_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        created_at = isoparse(d.pop("createdAt"))

        id = d.pop("id")

        name = d.pop("name")

        updated_at = isoparse(d.pop("updatedAt"))

        value = d.pop("value")

        color = d.pop("color", UNSET)

        parent_id = d.pop("parentId", UNSET)

        tag_response_dto = cls(
            created_at=created_at,
            id=id,
            name=name,
            updated_at=updated_at,
            value=value,
            color=color,
            parent_id=parent_id,
        )

        tag_response_dto.additional_properties = d
        return tag_response_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class Permission(str, Enum):
    ACTIVITY_CREATE = "activity.create"
    ACTIVITY_DELETE = "activity.delete"
    ACTIVITY_READ = "activity.read"
    ACTIVITY_STATISTICS = "activity.statistics"
    ACTIVITY_UPDATE = "activity.update"
    ADMINAUTH_UNLINKALL = "adminAuth.unlinkAll"
    ADMINUSER_CREATE = "adminUser.create"
    ADMINUSER_DELETE = "adminUser.delete"
    ADMINUSER_READ = "adminUser.read"
    ADMINUSER_UPDATE = "adminUser.update"
    ALBUMASSET_CREATE = "albumAsset.create"
    ALBUMASSET_DELETE = "albumAsset.delete"
    ALBUMUSER_CREATE = "albumUser.create"
    ALBUMUSER_DELETE = "albumUser.delete"
    ALBUMUSER_UPDATE = "albumUser.update"
    ALBUM_CREATE = "album.create"
    ALBUM_DELETE = "album.delete"
    ALBUM_DOWNLOAD = "album.download"
    ALBUM_READ = "album.read"
    ALBUM_SHARE = "album.share"
    ALBUM_STATISTICS = "album.statistics"
    ALBUM_UPDATE = "album.update"
    ALL = "all"
    APIKEY_CREATE = "apiKey.create"
    APIKEY_DELETE = "apiKey.delete"
    APIKEY_READ = "apiKey.read"
    APIKEY_UPDATE = "apiKey.update"
    ARCHIVE_READ = "archive.read"
    ASSET_DELETE = "asset.delete"
    ASSET_DOWNLOAD = "asset.download"
    ASSET_READ = "asset.read"
    ASSET_REPLACE = "asset.replace"
    ASSET_SHARE = "asset.share"
    ASSET_STATISTICS = "asset.statistics"
    ASSET_UPDATE = "asset.update"
    ASSET_UPLOAD = "asset.upload"
    ASSET_VIEW = "asset.view"
    AUTHDEVICE_DELETE = "authDevice.delete"
    AUTH_CHANGEPASSWORD = "auth.changePassword"
    DUPLICATE_DELETE = "duplicate.delete"
    DUPLICATE_READ = "duplicate.read"
    FACE_CREATE = "face.create"
    FACE_DELETE = "face.delete"
    FACE_READ = "face.read"
    FACE_UPDATE = "face.update"
    JOB_CREATE = "job.create"
    JOB_READ = "job.read"
    LIBRARY_CREATE = "library.create"
    LIBRARY_DELETE = "library.delete"
    LIBRARY_READ = "library.read"
    LIBRARY_STATISTICS = "library.statistics"
    LIBRARY_UPDATE = "library.update"
    MEMORYASSET_CREATE = "memoryAsset.create"
    MEMORYASSET_DELETE = "memoryAsset.delete"
    MEMORY_CREATE = "memory.create"
    MEMORY_DELETE = "memory.delete"
    MEMORY_READ = "memory.read"
    MEMORY_STATISTICS = "memory.statistics"
    MEMORY_UPDATE = "memory.update"
    NOTIFICATION_CREATE = "notification.create"
    NOTIFICATION_DELETE = "notification.delete"
    NOTIFICATION_READ = "notification.read"
    NOTIFICATION_UPDATE = "notification.update"
    PARTNER_CREATE = "partner.create"
    PARTNER_DELETE = "partner.delete"
    PARTNER_READ = "partner.read"
    PARTNER_UPDATE = "partner.update"
    PERSON_CREATE = "person.create"
    PERSON_DELETE = "person.delete"
    PERSON_MERGE = "person.merge"
    PERSON_READ = "person.read"
    PERSON_REASSIGN = "person.reassign"
    PERSON_STATISTICS = "person.statistics"
    PERSON_UPDATE = "person.update"
    PINCODE_CREATE = "pinCode.create"
    PINCODE_DELETE = "pinCode.delete"
    PINCODE_UPDATE = "pinCode.update"
    SERVERLICENSE_DELETE = "serverLicense.delete"
    SERVERLICENSE_READ = "serverLicense.read"
    SERVERLICENSE_UPDATE = "serverLicense.update"
    SERVER_ABOUT = "server.about"
    SERVER_APKLINKS = "server.apkLinks"
    SERVER_STATISTICS = "server.statistics"
    SERVER_STORAGE = "server.storage"
    SERVER_VERSIONCHECK = "server.versionCheck"
    SESSION_CREATE = "session.create"
    SESSION_DELETE = "session.delete"
    SESSION_LOCK = "session.lock"
    SESSION_READ = "session.read"
    SESSION_UPDATE = "session.update"
    SHAREDLINK_CREATE = "sharedLink.create"
    SHAREDLINK_DELETE = "sharedLink.delete"
    SHAREDLINK_READ = "sharedLink.read"
    SHAREDLINK_UPDATE = "sharedLink.update"
    STACK_CREATE = "stack.create"
    STACK_DELETE = "stack.delete"
    STACK_READ = "stack.read"
    STACK_UPDATE = "stack.update"
    SYNCCHECKPOINT_DELETE = "syncCheckpoint.delete"
    SYNCCHECKPOINT_READ = "syncCheckpoint.read"
    SYNCCHECKPOINT_UPDATE = "syncCheckpoint.update"
    SYNC_STREAM = "sync.stream"
    SYSTEMCONFIG_READ = "systemConfig.read"
    SYSTEMCONFIG_UPDATE = "systemConfig.update"
    SYSTEMMETADATA_READ = "systemMetadata.read"
    SYSTEMMETADATA_UPDATE = "systemMetadata.update"
    TAG_ASSET = "tag.asset"
    TAG_CREATE = "tag.create"
    TAG_DELETE = "tag.delete"
    TAG_READ = "tag.read"
    TAG_UPDATE = "tag.update"
    TIMELINE_DOWNLOAD = "timeline.download"
    TIMELINE_READ = "timeline.read"
    USERLICENSE_CREATE = "userLicense.create"
    USERLICENSE_DELETE = "userLicense.delete"
    USERLICENSE_READ = "userLicense.read"
    USERLICENSE_UPDATE = "userLicense.update"
    USERONBOARDING_DELETE = "userOnboarding.delete"
    USERONBOARDING_READ = "userOnboarding.read"
    USERONBOARDING_UPDATE = "userOnboarding.update"
    USERPREFERENCE_READ = "userPreference.read"
    USERPREFERENCE_UPDATE = "userPreference.update"
    USERPROFILEIMAGE_CREATE = "userProfileImage.create"
    USERPROFILEIMAGE_DELETE = "userProfileImage.delete"
    USERPROFILEIMAGE_READ = "userProfileImage.read"
    USERPROFILEIMAGE_UPDATE = "userProfileImage.update"
    USER_READ = "user.read"
    USER_UPDATE = "user.update"

    def __str__(self) -> str:
        return str(self.value)



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="ServerConfigDto")


@_attrs_define
class ServerConfigDto:
    """
    Attributes:
        external_domain (str):
        is_initialized (bool):
        is_onboarded (bool):
        login_page_message (str):
        map_dark_style_url (str):
        map_light_style_url (str):
        oauth_button_text (str):
        public_users (bool):
        trash_days (int):
        user_delete_delay (int):
    """

    external_domain: str
    is_initialized: bool
    is_onboarded: bool
    login_page_message: str
    map_dark_style_url: str
    map_light_style_url: str
    oauth_button_text: str
    public_users: bool
    trash_days: int
    user_delete_delay: int
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        external_domain = self.external_domain

        is_initialized = self.is_initialized

        is_onboarded = self.is_onboarded

        login_page_message = self.login_page_message

        map_dark_style_url = self.map_dark_style_url

        map_light_style_url = self.map_light_style_url

        oauth_button_text = self.oauth_button_text

        public_users = self.public_users

        trash_days = self.trash_days

        user_delete_delay = self.user_delete_delay

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "externalDomain": external_domain,
                "isInitialized": is_initialized,
                "isOnboarded": is_onboarded,
                "loginPageMessage": login_page_message,
                "mapDarkStyleUrl": map_dark_style_url,
                "mapLightStyleUrl": map_light_style_url,
                "oauthButtonText": oauth_button_text,
                "publicUsers": public_users,
                "trashDays": trash_days,
                "userDeleteDelay": user_delete_delay,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        external_domain = d.pop("externalDomain")

        is_initialized = d.pop("isInitialized")

        is_onboarded = d.pop("isOnboarded")

        login_page_message = d.pop("loginPageMessage")

        map_dark_style_url = d.pop("mapDarkStyleUrl")

        map_light_style_url = d.pop("mapLightStyleUrl")

        oauth_button_text = d.pop("oauthButtonText")

        public_users = d.pop("publicUsers")

        trash_days = d.pop("trashDays")

        user_delete_delay = d.pop("userDeleteDelay")

        server_config_dto = cls(
            external_domain=external_domain,
            is_initialized=is_initialized,
            is_onboarded=is_onboarded,
            login_page_message=login_page_message,
            map_dark_style_url=map_dark_style_url,
            map_light_style_url=map_light_style_url,
            oauth_button_text=oauth_button_text,
            public_users=public_users,
            trash_days=trash_days,
            user_delete_delay=user_delete_delay,
        )

        server_config_dto.additional_properties = d
        return server_config_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.asset_type_enum import AssetTypeEnum
from ..models.asset_visibility import AssetVisibility

T = TypeVar("T", bound="SyncAssetV1")


@_attrs_define
class SyncAssetV1:
    """
    Attributes:
        checksum (str):
        deleted_at (Union[None, datetime.datetime]):
        duration (Union[None, str]):
        file_created_at (Union[None, datetime.datetime]):
        file_modified_at (Union[None, datetime.datetime]):
        id (str):
        is_favorite (bool):
        library_id (Union[None, str]):
        live_photo_video_id (Union[None, str]):
        local_date_time (Union[None, datetime.datetime]):
        original_file_name (str):
        owner_id (str):
        stack_id (Union[None, str]):
        thumbhash (Union[None, str]):
        type_ (AssetTypeEnum):
        visibility (AssetVisibility):
    """

    checksum: str
    deleted_at: Union[None, datetime.datetime]
    duration: Union[None, str]
    file_created_at: Union[None, datetime.datetime]
    file_modified_at: Union[None, datetime.datetime]
    id: str
    is_favorite: bool
    library_id: Union[None, str]
    live_photo_video_id: Union[None, str]
    local_date_time: Union[None, datetime.datetime]
    original_file_name: str
    owner_id: str
    stack_id: Union[None, str]
    thumbhash: Union[None, str]
    type_: AssetTypeEnum
    visibility: AssetVisibility
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        checksum = self.checksum

        deleted_at: Union[None, str]
        if isinstance(self.deleted_at, datetime.datetime):
            deleted_at = self.deleted_at.isoformat()
        else:
            deleted_at = self.deleted_at

        duration: Union[None, str]
        duration = self.duration

        file_created_at: Union[None, str]
        if isinstance(self.file_created_at, datetime.datetime):
            file_created_at = self.file_created_at.isoformat()
        else:
            file_created_at = self.file_created_at

        file_modified_at: Union[None, str]
        if isinstance(self.file_modified_at, datetime.datetime):
            file_modified_at = self.file_modified_at.isoformat()
        else:
            file_modified_at = self.file_modified_at

        id = self.id

        is_favorite = self.is_favorite

        library_id: Union[None, str]
        library_id = self.library_id

        live_photo_video_id: Union[None, str]
        live_photo_video_id = self.live_photo_video_id

        local_date_time: Union[None, str]
        if isinstance(self.local_date_time, datetime.datetime):
            local_date_time = self.local_date_time.isoformat()
        else:
            local_date_time = self.local_date_time

        original_file_name = self.original_file_name

        owner_id = self.owner_id

        stack_id: Union[None, str]
        stack_id = self.stack_id

        thumbhash: Union[None, str]
        thumbhash = self.thumbhash

        type_ = self.type_.value

        visibility = self.visibility.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "checksum": checksum,
                "deletedAt": deleted_at,
                "duration": duration,
                "fileCreatedAt": file_created_at,
                "fileModifiedAt": file_modified_at,
                "id": id,
                "isFavorite": is_favorite,
                "libraryId": library_id,
                "livePhotoVideoId": live_photo_video_id,
                "localDateTime": local_date_time,
                "originalFileName": original_file_name,
                "ownerId": owner_id,
                "stackId": stack_id,
                "thumbhash": thumbhash,
                "type": type_,
                "visibility": visibility,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        checksum = d.pop("checksum")

        def _parse_deleted_at(data: object) -> Union[None, datetime.datetime]:
            if data is None:
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                deleted_at_type_0 = isoparse(data)

                return deleted_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, datetime.datetime], data)

        deleted_at = _parse_deleted_at(d.pop("deletedAt"))

        def _parse_duration(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        duration = _parse_duration(d.pop("duration"))

        def _parse_file_created_at(data: object) -> Union[None, datetime.datetime]:
            if data is None:
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                file_created_at_type_0 = isoparse(data)

                return file_created_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, datetime.datetime], data)

        file_created_at = _parse_file_created_at(d.pop("fileCreatedAt"))

        def _parse_file_modified_at(data: object) -> Union[None, datetime.datetime]:
            if data is None:
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                file_modified_at_type_0 = isoparse(data)

                return file_modified_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, datetime.datetime], data)

        file_modified_at = _parse_file_modified_at(d.pop("fileModifiedAt"))

        id = d.pop("id")

        is_favorite = d.pop("isFavorite")

        def _parse_library_id(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        library_id = _parse_library_id(d.pop("libraryId"))

        def _parse_live_photo_video_id(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        live_photo_video_id = _parse_live_photo_video_id(d.pop("livePhotoVideoId"))

        def _parse_local_date_time(data: object) -> Union[None, datetime.datetime]:
            if data is None:
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                local_date_time_type_0 = isoparse(data)

                return local_date_time_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, datetime.datetime], data)

        local_date_time = _parse_local_date_time(d.pop("localDateTime"))

        original_file_name = d.pop("originalFileName")

        owner_id = d.pop("ownerId")

        def _parse_stack_id(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        stack_id = _parse_stack_id(d.pop("stackId"))

        def _parse_thumbhash(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        thumbhash = _parse_thumbhash(d.pop("thumbhash"))

        type_ = AssetTypeEnum(d.pop("type"))

        visibility = AssetVisibility(d.pop("visibility"))

        sync_asset_v1 = cls(
            checksum=checksum,
            deleted_at=deleted_at,
            duration=duration,
            file_created_at=file_created_at,
            file_modified_at=file_modified_at,
            id=id,
            is_favorite=is_favorite,
            library_id=library_id,
            live_photo_video_id=live_photo_video_id,
            local_date_time=local_date_time,
            original_file_name=original_file_name,
            owner_id=owner_id,
            stack_id=stack_id,
            thumbhash=thumbhash,
            type_=type_,
            visibility=visibility,
        )

        sync_asset_v1.additional_properties = d
        return sync_asset_v1

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="TagCreateDto")


@_attrs_define
class TagCreateDto:
    """
    Attributes:
        name (str):
        color (Union[Unset, str]):
        parent_id (Union[None, UUID, Unset]):
    """

    name: str
    color: Union[Unset, str] = UNSET
    parent_id: Union[None, UUID, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        color = self.color

        parent_id: Union[None, Unset, str]
        if isinstance(self.parent_id, Unset):
            parent_id = UNSET
        elif isinstance(self.parent_id, UUID):
            parent_id = str(self.parent_id)
        else:
            parent_id = self.parent_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
            }
        )
        if color is not UNSET:
            field_dict["color"] = color
        if parent_id is not UNSET:
            field_dict["parentId"] = parent_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        name = d.pop("name")

        color = d.pop("color", UNSET)

        def _parse_parent_id(data: object) -> Union[None, UUID, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                parent_id_type_0 = UUID(data)

                return parent_id_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, UUID, Unset], data)

        parent_id = _parse_parent_id(d.pop("parentId", UNSET))

        tag_create_dto = cls(
            name=name,
            color=color,
            parent_id=parent_id,
        )

        tag_create_dto.additional_properties = d
        return tag_create_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.asset_bulk_upload_check_item import AssetBulkUploadCheckItem


T = TypeVar("T", bound="AssetBulkUploadCheckDto")


@_attrs_define
class AssetBulkUploadCheckDto:
    """
    Attributes:
        assets (list['AssetBulkUploadCheckItem']):
    """

    assets: list["AssetBulkUploadCheckItem"]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        assets = []
        for assets_item_data in self.assets:
            assets_item = assets_item_data.to_dict()
            assets.append(assets_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "assets": assets,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.asset_bulk_upload_check_item import AssetBulkUploadCheckItem

        d = dict(src_dict)
        assets = []
        _assets = d.pop("assets")
        for assets_item_data in _assets:
            assets_item = AssetBulkUploadCheckItem.from_dict(assets_item_data)

            assets.append(assets_item)

        asset_bulk_upload_check_dto = cls(
            assets=assets,
        )

        asset_bulk_upload_check_dto.additional_properties = d
        return asset_bulk_upload_check_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.albums_update import AlbumsUpdate
    from ..models.avatar_update import AvatarUpdate
    from ..models.cast_update import CastUpdate
    from ..models.download_update import DownloadUpdate
    from ..models.email_notifications_update import EmailNotificationsUpdate
    from ..models.folders_update import FoldersUpdate
    from ..models.memories_update import MemoriesUpdate
    from ..models.people_update import PeopleUpdate
    from ..models.purchase_update import PurchaseUpdate
    from ..models.ratings_update import RatingsUpdate
    from ..models.shared_links_update import SharedLinksUpdate
    from ..models.tags_update import TagsUpdate


T = TypeVar("T", bound="UserPreferencesUpdateDto")


@_attrs_define
class UserPreferencesUpdateDto:
    """
    Attributes:
        albums (Union[Unset, AlbumsUpdate]):
        avatar (Union[Unset, AvatarUpdate]):
        cast (Union[Unset, CastUpdate]):
        download (Union[Unset, DownloadUpdate]):
        email_notifications (Union[Unset, EmailNotificationsUpdate]):
        folders (Union[Unset, FoldersUpdate]):
        memories (Union[Unset, MemoriesUpdate]):
        people (Union[Unset, PeopleUpdate]):
        purchase (Union[Unset, PurchaseUpdate]):
        ratings (Union[Unset, RatingsUpdate]):
        shared_links (Union[Unset, SharedLinksUpdate]):
        tags (Union[Unset, TagsUpdate]):
    """

    albums: Union[Unset, "AlbumsUpdate"] = UNSET
    avatar: Union[Unset, "AvatarUpdate"] = UNSET
    cast: Union[Unset, "CastUpdate"] = UNSET
    download: Union[Unset, "DownloadUpdate"] = UNSET
    email_notifications: Union[Unset, "EmailNotificationsUpdate"] = UNSET
    folders: Union[Unset, "FoldersUpdate"] = UNSET
    memories: Union[Unset, "MemoriesUpdate"] = UNSET
    people: Union[Unset, "PeopleUpdate"] = UNSET
    purchase: Union[Unset, "PurchaseUpdate"] = UNSET
    ratings: Union[Unset, "RatingsUpdate"] = UNSET
    shared_links: Union[Unset, "SharedLinksUpdate"] = UNSET
    tags: Union[Unset, "TagsUpdate"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        albums: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.albums, Unset):
            albums = self.albums.to_dict()

        avatar: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.avatar, Unset):
            avatar = self.avatar.to_dict()

        cast: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.cast, Unset):
            cast = self.cast.to_dict()

        download: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.download, Unset):
            download = self.download.to_dict()

        email_notifications: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.email_notifications, Unset):
            email_notifications = self.email_notifications.to_dict()

        folders: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.folders, Unset):
            folders = self.folders.to_dict()

        memories: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.memories, Unset):
            memories = self.memories.to_dict()

        people: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.people, Unset):
            people = self.people.to_dict()

        purchase: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.purchase, Unset):
            purchase = self.purchase.to_dict()

        ratings: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.ratings, Unset):
            ratings = self.ratings.to_dict()

        shared_links: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.shared_links, Unset):
            shared_links = self.shared_links.to_dict()

        tags: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if albums is not UNSET:
            field_dict["albums"] = albums
        if avatar is not UNSET:
            field_dict["avatar"] = avatar
        if cast is not UNSET:
            field_dict["cast"] = cast
        if download is not UNSET:
            field_dict["download"] = download
        if email_notifications is not UNSET:
            field_dict["emailNotifications"] = email_notifications
        if folders is not UNSET:
            field_dict["folders"] = folders
        if memories is not UNSET:
            field_dict["memories"] = memories
        if people is not UNSET:
            field_dict["people"] = people
        if purchase is not UNSET:
            field_dict["purchase"] = purchase
        if ratings is not UNSET:
            field_dict["ratings"] = ratings
        if shared_links is not UNSET:
            field_dict["sharedLinks"] = shared_links
        if tags is not UNSET:
            field_dict["tags"] = tags

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.albums_update import AlbumsUpdate
        from ..models.avatar_update import AvatarUpdate
        from ..models.cast_update import CastUpdate
        from ..models.download_update import DownloadUpdate
        from ..models.email_notifications_update import EmailNotificationsUpdate
        from ..models.folders_update import FoldersUpdate
        from ..models.memories_update import MemoriesUpdate
        from ..models.people_update import PeopleUpdate
        from ..models.purchase_update import PurchaseUpdate
        from ..models.ratings_update import RatingsUpdate
        from ..models.shared_links_update import SharedLinksUpdate
        from ..models.tags_update import TagsUpdate

        d = dict(src_dict)
        _albums = d.pop("albums", UNSET)
        albums: Union[Unset, AlbumsUpdate]
        if isinstance(_albums, Unset):
            albums = UNSET
        else:
            albums = AlbumsUpdate.from_dict(_albums)

        _avatar = d.pop("avatar", UNSET)
        avatar: Union[Unset, AvatarUpdate]
        if isinstance(_avatar, Unset):
            avatar = UNSET
        else:
            avatar = AvatarUpdate.from_dict(_avatar)

        _cast = d.pop("cast", UNSET)
        cast: Union[Unset, CastUpdate]
        if isinstance(_cast, Unset):
            cast = UNSET
        else:
            cast = CastUpdate.from_dict(_cast)

        _download = d.pop("download", UNSET)
        download: Union[Unset, DownloadUpdate]
        if isinstance(_download, Unset):
            download = UNSET
        else:
            download = DownloadUpdate.from_dict(_download)

        _email_notifications = d.pop("emailNotifications", UNSET)
        email_notifications: Union[Unset, EmailNotificationsUpdate]
        if isinstance(_email_notifications, Unset):
            email_notifications = UNSET
        else:
            email_notifications = EmailNotificationsUpdate.from_dict(
                _email_notifications
            )

        _folders = d.pop("folders", UNSET)
        folders: Union[Unset, FoldersUpdate]
        if isinstance(_folders, Unset):
            folders = UNSET
        else:
            folders = FoldersUpdate.from_dict(_folders)

        _memories = d.pop("memories", UNSET)
        memories: Union[Unset, MemoriesUpdate]
        if isinstance(_memories, Unset):
            memories = UNSET
        else:
            memories = MemoriesUpdate.from_dict(_memories)

        _people = d.pop("people", UNSET)
        people: Union[Unset, PeopleUpdate]
        if isinstance(_people, Unset):
            people = UNSET
        else:
            people = PeopleUpdate.from_dict(_people)

        _purchase = d.pop("purchase", UNSET)
        purchase: Union[Unset, PurchaseUpdate]
        if isinstance(_purchase, Unset):
            purchase = UNSET
        else:
            purchase = PurchaseUpdate.from_dict(_purchase)

        _ratings = d.pop("ratings", UNSET)
        ratings: Union[Unset, RatingsUpdate]
        if isinstance(_ratings, Unset):
            ratings = UNSET
        else:
            ratings = RatingsUpdate.from_dict(_ratings)

        _shared_links = d.pop("sharedLinks", UNSET)
        shared_links: Union[Unset, SharedLinksUpdate]
        if isinstance(_shared_links, Unset):
            shared_links = UNSET
        else:
            shared_links = SharedLinksUpdate.from_dict(_shared_links)

        _tags = d.pop("tags", UNSET)
        tags: Union[Unset, TagsUpdate]
        if isinstance(_tags, Unset):
            tags = UNSET
        else:
            tags = TagsUpdate.from_dict(_tags)

        user_preferences_update_dto = cls(
            albums=albums,
            avatar=avatar,
            cast=cast,
            download=download,
            email_notifications=email_notifications,
            folders=folders,
            memories=memories,
            people=people,
            purchase=purchase,
            ratings=ratings,
            shared_links=shared_links,
            tags=tags,
        )

        user_preferences_update_dto.additional_properties = d
        return user_preferences_update_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="ValidateLibraryImportPathResponseDto")


@_attrs_define
class ValidateLibraryImportPathResponseDto:
    """
    Attributes:
        import_path (str):
        is_valid (bool):  Default: False.
        message (Union[Unset, str]):
    """

    import_path: str
    is_valid: bool = False
    message: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        import_path = self.import_path

        is_valid = self.is_valid

        message = self.message

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "importPath": import_path,
                "isValid": is_valid,
            }
        )
        if message is not UNSET:
            field_dict["message"] = message

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        import_path = d.pop("importPath")

        is_valid = d.pop("isValid")

        message = d.pop("message", UNSET)

        validate_library_import_path_response_dto = cls(
            import_path=import_path,
            is_valid=is_valid,
            message=message,
        )

        validate_library_import_path_response_dto.additional_properties = d
        return validate_library_import_path_response_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ToneMapping(str, Enum):
    DISABLED = "disabled"
    HABLE = "hable"
    MOBIUS = "mobius"
    REINHARD = "reinhard"

    def __str__(self) -> str:
        return str(self.value)



from collections.abc import Mapping
from typing import Any, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="AlbumStatisticsResponseDto")


@_attrs_define
class AlbumStatisticsResponseDto:
    """
    Attributes:
        not_shared (int):
        owned (int):
        shared (int):
    """

    not_shared: int
    owned: int
    shared: int
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        not_shared = self.not_shared

        owned = self.owned

        shared = self.shared

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "notShared": not_shared,
                "owned": owned,
                "shared": shared,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        not_shared = d.pop("notShared")

        owned = d.pop("owned")

        shared = d.pop("shared")

        album_statistics_response_dto = cls(
            not_shared=not_shared,
            owned=owned,
            shared=shared,
        )

        album_statistics_response_dto.additional_properties = d
        return album_statistics_response_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Immich CLI commands."""

import time
import click
from uuid import UUID
import sys
from rich.progress import Progress, SpinnerColumn, BarColumn
import functools
from loguru import logger
from immichporter.commands import logging_options
from rich.console import Console
from rich.table import Table
from rich.progress import (
    TextColumn,
    TimeRemainingColumn,
    TimeElapsedColumn,
)

from immichporter.utils import generate_password
from datetime import datetime, timezone
from immichporter.immich.immich import ImmichClient, immich_api_client
from immichporter.database import (
    get_db_session,
    get_photos_from_db,
    get_albums_without_immich_id,
    get_users,
)
from immichporter.immich.client.models import AlbumUserCreateDto, AlbumUserRole
from immichporter.immich.client.api.users import get_my_user
from immichporter.models import Photo
from immichporter.immich.db import ImmichDBClient
from immichporter.database import get_db_session as get_sqlite_session
from immichporter.models import User
from immichporter.immich.client.models import (
    JobName,
)


def format_time(seconds: float) -> str:
    """Format time in seconds to a human-readable string (e.g., '1h 23m 45s' or '5m 23s')."""
    if seconds is None or seconds < 0:
        return "Calculating..."
    hours, remainder = divmod(int(seconds), 3600)
    minutes, seconds = divmod(remainder, 60)
    if hours > 0:
        return f"{hours}h {minutes:02d}m {seconds:02d}s"
    return f"{minutes:02d}m {seconds:02d}s"


console = Console()

# Create a Click command group
cli_immich = click.Group("immich", help="Update users and albums in immich")


def immich_options(f):
    """Common options for Immich commands."""

    @click.option(
        "--endpoint",
        envvar="IMMICH_ENDPOINT",
        default="http://localhost:2283",
        help="Immich server URL (default: http://localhost:2283)",
        show_default=True,
        show_envvar=True,
    )
    @click.option(
        "-k",
        "--api-key",
        envvar="IMMICH_API_KEY",
        required=True,
        help="Immich API key",
        show_envvar=True,
    )
    @click.option(
        "--insecure",
        is_flag=True,
        envvar="IMMICH_INSECURE",
        help="Skip SSL certificate verification",
        show_envvar=True,
    )
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        endpoint = kwargs["endpoint"]
        api_key = kwargs["api_key"]
        insecure = kwargs["insecure"]
        client_api = immich_api_client(
            endpoint=endpoint, api_key=api_key, insecure=insecure
        )
        kwargs["immich_api"] = client_api
        kwargs["immich"] = ImmichClient(client=client_api)
        return f(*args, **kwargs)

    return wrapper


# Add commands to the immich group
@cli_immich.command()
@immich_options
@click.option(
    "--limit", type=int, default=50, help="Maximum number of albums to return"
)
@click.option("--shared/--no-shared", default=None, help="Filter by shared status")
@logging_options
def list_albums(immich: ImmichClient, limit, shared, **options):
    """List all albums on the Immich server."""
    logger.info(f"Fetching albums from '{immich.endpoint}'")

    try:
        albums = immich.get_albums(limit=limit, shared=shared)
        logger.info(f"Retrieved {len(albums) if albums else 0} albums")

        if not albums:
            console.print("[red]No albums found.[/]")
            return

        # Create a table
        table = Table(show_header=True, header_style="bold magenta")
        table.add_column("ID", style="dim", width=36)
        table.add_column("Album Name")
        table.add_column("Asset Count")
        table.add_column("Shared")
        table.add_column("Created At")

        for album in albums:
            table.add_row(
                str(album.id),
                album.album_name,
                str(album.asset_count),
                "✓" if album.shared else "✗",
                album.created_at.strftime("%Y-%m-%d %H:%M")
                if hasattr(album, "created_at")
                else "N/A",
            )

        # Render the table
        console.print(table)
    except Exception as e:
        logger.error(f"Error fetching albums: {str(e)}")


@cli_immich.command()
@immich_options
@click.option(
    "--dry-run",
    is_flag=True,
    help="Run without making any changes",
    default=False,
)
@logging_options
def update_photos(immich: ImmichClient, dry_run: bool, **options):
    """Update our photos with their Immich IDs by searching for them in Immich."""
    logger.info("Starting photo update process")

    # Get photos without immich_id
    session = get_db_session()
    photos = get_photos_from_db(session, has_immich_id=False)

    if not photos:
        logger.info("All photos already have immich_ids")
        return

    if dry_run:
        console.print("[yellow]\n DRY RUN MODE - No changes will be made\n[/yellow]")
        console.print("[yellow]\n🚧 DRY RUN MODE - No changes will be made\n[/yellow]")

    logger.info(f"Found {len(photos)} photos without immich_id")

    from concurrent.futures import ThreadPoolExecutor, as_completed

    # Initialize counters
    updated_count = 0
    not_found_count = 0

    # Configure progress bar with total count and improved time display
    progress_columns = [
        TextColumn("Updating photos", style="white"),
        BarColumn(
            bar_width=None,  # Will use full width
            complete_style="blue",
            finished_style="green",
            pulse_style="white",
        ),
        TextColumn("[white]{task.percentage:>3.0f}%[/white]", justify="right"),
        TextColumn("•", style="white"),
        TextColumn("[white]{task.completed}[/]/[blue]{task.total}[/]", justify="right"),
        TextColumn("[red]({task.fields[not_found]} not found)[/red]", justify="left"),
        TextColumn("•", style="white"),
        TimeElapsedColumn(),
        TextColumn("(ETA:", style="white"),
        TimeRemainingColumn(compact=True),
        TextColumn(")", style="white"),
    ]

    def process_photo(photo_id, filename, date_taken, immich, dry_run):
        """Process a single photo and return the result.

        Args:
            photo_id: The ID of the photo in the database
            filename: The filename of the photo
            date_taken: The date the photo was taken (optional)
            immich: The Immich client instance
            dry_run: Whether this is a dry run

        Returns:
            tuple: (photo_id, matched_asset_id, error_message)
        """
        try:
            # First try with both filename and date
            results = immich.search_assets(filename=filename, taken=date_taken)

            # If no results, try with just the filename
            if not results:
                results = immich.search_assets(filename=filename)

            if results:
                return photo_id, results[0].id, None  # Return first match ID
            else:
                return photo_id, None, "not_found"

        except Exception as e:
            return photo_id, None, str(e)

    with Progress(
        *progress_columns,
        console=console,
        transient=True,
        refresh_per_second=10,
        speed_estimate_period=30.0,
    ) as progress:
        # Initialize task with proper time formatting
        task = progress.add_task(
            "[cyan]Updating photos...", total=len(photos), completed=0, not_found=0
        )

        # Process photos in batches
        BATCH_SIZE = 20
        MAX_WORKERS = 5

        for i in range(0, len(photos), BATCH_SIZE):
            batch = photos[i : i + BATCH_SIZE]
            batch_updates = []

            with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
                # Submit all photos in the current batch with just the data we need
                future_to_photo_id = {
                    executor.submit(
                        process_photo,
                        photo.id,
                        photo.filename,
                        photo.date_taken if hasattr(photo, "date_taken") else None,
                        immich,
                        dry_run,
                    ): photo.id
                    for photo in batch
                }

                # Process completed futures as they complete
                for future in as_completed(future_to_photo_id):
                    photo_id, matched_asset_id, error = future.result()

                    # Get the photo object from the database to ensure it's fresh
                    try:
                        photo = session.get(Photo, photo_id)
                        if not photo:
                            logger.warning(
                                f"Photo with ID {photo_id} not found in database"
                            )
                            not_found_count += 1
                            continue

                        if error == "not_found":
                            logger.info(f"No match found for {photo.filename}")
                            not_found_count += 1
                        elif error:
                            logger.error(f"Error processing {photo.filename}: {error}")
                            continue
                        else:
                            # Add to batch updates
                            batch_updates.append(
                                {
                                    "id": photo.id,
                                    "immich_id": matched_asset_id,
                                    "updated_at": datetime.now(timezone.utc),
                                }
                            )
                            updated_count += 1

                    except Exception as e:
                        logger.error(f"Error updating photo {photo_id}: {str(e)}")
                        continue

                    # Update progress
                    progress.update(
                        task,
                        advance=1,
                        completed=min(updated_count + not_found_count, len(photos)),
                        not_found=not_found_count,
                    )

            # Process batch updates
            if batch_updates:
                if not dry_run:
                    try:
                        # Update all photos in the batch at once
                        session.bulk_update_mappings(Photo, batch_updates)  # type: ignore
                        session.commit()
                        logger.debug(f"Updated {len(batch_updates)} photos in batch")
                    except Exception as e:
                        session.rollback()
                        logger.error(f"Error updating batch: {str(e)}")
                else:
                    logger.debug(
                        f"[DRY RUN] Would update {len(batch_updates)} photos in batch"
                    )

            # Small delay between batches to avoid overwhelming the server
            time.sleep(0.1)

    # Log summary
    logger.info(f"Update complete. Processed {len(photos)} photos:")
    logger.info(f"  • Successfully updated: {updated_count}")
    logger.info(f"  • Not found: {not_found_count}")

    if not dry_run:
        logger.success(f"Successfully updated {updated_count} photos in Immich")
    else:
        logger.info(f"Dry run complete. Would update {updated_count} photos")


@cli_immich.command()
@immich_options
@click.option("-a", "--all", is_flag=True, help="Delete all albums")
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    help="Run without making any changes",
    default=False,
)
@logging_options
def delete_albums(immich: ImmichClient, all: bool, dry_run: bool, **options):
    """Delete albums from Immich."""
    logger.info("Starting album deletion process")
    albums = immich.get_albums()
    if all:
        for album in albums:
            if not dry_run:
                immich.delete_album(album.id)
            else:
                console.print(
                    f"[yellow]DRY RUN:[/] Would delete album: [blue]{album.album_name}[/]"
                )


@cli_immich.command()
@immich_options
@click.option(
    "--limit", type=int, default=None, help="Limit the number of albums to sync"
)
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    help="Run without making any changes",
    default=False,
)
@logging_options
def sync_albums(immich: ImmichClient, limit: int | None, dry_run: bool, **options):
    """Sync albums from our database to Immich."""
    logger.info("Starting album sync process")

    session = get_db_session()

    # Get albums that haven't been synced to Immich yet
    albums = get_albums_without_immich_id(session)
    if limit:
        albums = albums[:limit]

    if not albums:
        logger.info("All albums already synced to Immich")
        return

    if dry_run:
        console.print("[yellow]\n🚧 DRY RUN MODE - No changes will be made\n[/yellow]")

    logger.info(f"Found {len(albums)} albums to sync")

    # Configure progress bar
    progress_columns = [
        TextColumn("Syncing albums", style="white"),
        BarColumn(bar_width=None, complete_style="blue", finished_style="green"),
        TextColumn("[white]{task.percentage:>3.0f}%[/white]", justify="right"),
        TextColumn("•", style="white"),
        TextColumn("[white]{task.completed}/{task.total}", justify="right"),
    ]

    my_user = get_my_user.sync(client=immich.client)
    assert my_user is not None
    if my_user is None:
        logger.error("Failed to get my user")
        sys.exit(1)

    with Progress(*progress_columns, console=console) as progress:
        task = progress.add_task("Syncing albums...", total=len(albums))

        for album in albums:
            try:
                # Get album users with immich_id
                album_users = [
                    AlbumUserCreateDto(
                        user_id=UUID(user.immich_user_id), role=AlbumUserRole.VIEWER
                    )
                    for user in album.users
                    if user.immich_user_id is not None
                    and user.immich_user_id != my_user.id
                ]
                album_user_names = [
                    user.immich_name for user in album.users if user.immich_name
                ]

                # Get album photos with immich_id
                photos = get_photos_from_db(
                    session, album_id=album.id, has_immich_id=True
                )
                photo_ids = [photo.immich_id for photo in photos]
                album_name = (
                    album.immich_title if album.immich_title else album.source_title
                )

                if dry_run:
                    progress.console.print("[yellow][DRY RUN][/]")
                progress.console.print(
                    f"Create album: [blue]{album_name}[/] with [blue]{len(photo_ids)}[/] photos"
                )
                if album_user_names:
                    progress.console.print(f"[dim]{', '.join(album_user_names)}[/]")
                if not dry_run:
                    # Create album in Immich
                    immich_album = immich.create_album(
                        name=album_name,
                        description=None,
                        users=album_users,
                        assets=photo_ids,
                    )

                    # Update album with immich_id
                    album.immich_id = immich_album.id
                    session.add(album)
                    session.commit()
                else:
                    time.sleep(0.1)

            except Exception as e:
                logger.error(f"Error syncing album {album.source_title}: {str(e)}")
                session.rollback()
            finally:
                progress.update(task, advance=1)

    logger.info("Album sync complete")


@cli_immich.command()
@immich_options
@click.option(
    "--dry-run",
    is_flag=True,
    help="Run without making any changes",
    default=False,
)
@logging_options
def update_users(immich: ImmichClient, dry_run: bool, **options):
    """Update our users with their Immich IDs by searching for them in Immich."""
    logger.info("Starting users update process")
    session = get_db_session()
    db_users = get_users(session)
    immich_users = immich.get_users()
    for db_user in db_users:
        if not db_user.add_to_immich:
            continue
        immich_name = db_user.immich_name
        immich_email = db_user.immich_email
        mail_fmt = f"{immich_email}"
        console.print(f"{immich_name:<20} [blue]{mail_fmt:<30}[/]", end="")
        set_immich_id = db_user.immich_user_id
        immich_id = set_immich_id
        if not immich_id:  # check for email
            for immich_user in immich_users:
                if immich_user.email == immich_email:
                    immich_id = immich_user.id
                    break
        if not immich_id:  # check for name
            for immich_user in immich_users:
                if immich_user.name == immich_name:
                    immich_id = immich_user.id
                    break
        if set_immich_id is None and immich_id:  # add to DB
            if not dry_run:
                console.print(f"add id: {immich_id}")
                db_user.immich_user_id = immich_id
                session.add(db_user)
                session.commit()
            else:
                console.print(f"[yellow][DRY RUN][/] add id: {immich_id}")
        elif set_immich_id is not None:
            console.print(f"[green]id already set: {immich_id}[/]")
        else:
            # create new user
            password = generate_password()
            if not dry_run:
                assert immich_name is not None
                assert immich_email is not None
                immich.add_user(
                    name=immich_name, email=immich_email, password=password, quota_gb=15
                )
                console.print("[red]add user to immich")
                db_user.immich_initial_password = password
                session.add(db_user)
                session.commit()
            else:
                console.print("[yellow][DRY RUN][/] add user to immich")


@cli_immich.command()
@immich_options
@click.option("--tag-name", help="Name of the tag")
@click.option("-m", "--remove-tag", is_flag=True, help="Remove tag name as well")
@click.option(
    "--all", is_flag=True, help="Untag all assets. '--tag-name' will be ignored"
)
@click.option(
    "-e",
    "--exclude",
    help="Name of tags to exclude if '--all' is used. Make sure to include parents as well!",
    multiple=True,
)
@click.option(
    "--dry-run/--no-dry-run",
    default=True,
    help="Run without making any changes",
    show_default=True,
)
@logging_options
def untag_assets(
    immich: ImmichClient,
    tag_name: str,
    remove_tag: bool,
    all: bool,
    exclude: list[str],
    dry_run: bool,
    **options,
):
    """Untag assets with a specific tag. Children tags will be removed as well."""
    if all:
        tags = immich.get_tags()
        tags = sorted(tags, key=lambda tag: len(tag.value.split("/")), reverse=True)
        tags = [tag for tag in tags if tag.name not in exclude]
        if not tags:
            logger.error("No tags found")
            sys.exit(1)
    else:
        tags = [immich.get_tags(filter_name=tag_name)[0]]
        if not tags:
            logger.error(f"Tag '{tag_name}' not found")
            sys.exit(1)
    for tag in tags:
        console.print(f"Untagging assets with tag '{tag.value}'")
        if dry_run:
            assets = immich.timeline_assets(tag_id=tag.id)
            console.print(
                f"[yellow][DRY RUN][/] Would delete tag '{tag.value}' from {len(list(assets))} assets (use '--no-dry-run' to actually remove)"
            )
            # sys.exit(0)
        else:
            rm = immich.untag_assets(tag=tag, remove_tag=remove_tag)
            console.print(f"Deleted tag '{tag.value}' from {rm} assets")


@cli_immich.command("adjust-owners")
@immich_options
@click.option(
    "--db-host",
    envvar="DB_HOST",
    default="database",
    help="PostgreSQL host. Can also be set via DB_HOST environment variable.",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--db-port",
    envvar="DB_PORT",
    default=5432,
    type=int,
    help="PostgreSQL port. Can also be set via DB_PORT environment variable.",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--db-username",
    envvar="DB_USERNAME",
    default="immich",
    help="PostgreSQL username. Can also be set via DB_USERNAME environment variable.",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--db-password",
    envvar="DB_PASSWORD",
    default="immich",
    help="PostgreSQL password. Can also be set via DB_PASSWORD environment variable.",
    show_envvar=True,
    show_default=False,
    hide_input=True,
)
@click.option(
    "--db-name",
    envvar="DB_NAME",
    default="immich",
    help="PostgreSQL database name. Can also be set via DB_NAME environment variable.",
    show_envvar=True,
    show_default=True,
)
@click.option(
    "--dry-run/--no-dry-run",
    is_flag=True,
    default=True,
    help="Show what would be changed without making any changes",
    show_default=True,
)
@click.option(
    "--batch-size",
    default=25,
    type=int,
    help="Number of assets to process in a batch",
    show_default=True,
)
@click.option(
    "--storage-migration/--no-storage-migration",
    is_flag=True,
    default=True,
    help="Run 'storage template migration' job after adjusting owners",
    show_default=True,
)
@click.option(
    "--db-backup/--no-db-backup",
    is_flag=True,
    default=True,
    help="Run 'storage template migration' job after adjusting owners",
    show_default=True,
)
@logging_options
def adjust_owners_command(
    immich: ImmichClient,
    db_host: str,
    db_port: int,
    db_username: str,
    db_password: str,
    db_name: str,
    dry_run: bool,
    batch_size: int,
    log_level: str,
    storage_migration: bool,
    db_backup: bool,
    **options,
):
    """Adjust asset ownership in Immich based on SQLite database."""
    log_level = options.get("log_level", "INFO").upper()

    if db_backup:
        console.print("[yellow]Running database backup...[/]")
        console.print("[dim]wait 2 minutes...[/]")
        immich.run_db_backup(180)

    # Construct PostgreSQL URL
    postgres_url = (
        f"postgresql://{db_username}:{db_password}@{db_host}:{db_port}/{db_name}"
    )

    if log_level == "DEBUG":
        console.print(
            f"[yellow]Using PostgreSQL URL: {postgres_url.split('@')[0]}@...[/]"
        )

    try:
        # Initialize Immich DB client
        immich_client = ImmichDBClient(postgres_url, echo=(log_level == "DEBUG"))
    except Exception as e:
        console.print(f"[red]Error connecting to Immich database: {e}[/]")
        return

    try:
        # Get SQLite session
        db_session = get_sqlite_session()

        # Get all users with add_to_immich=True
        users = db_session.query(User).filter(User.add_to_immich == True).all()  # noqa: E712
        if not users:
            console.print("[yellow]No users found with add_to_immich=True[/]")
            return

        console.print(f"Found {len(users)} users to process")

        for user in users:
            if not user.immich_user_id:
                console.print(
                    f"[yellow]Skipping user {user.source_name}: No immich_user_id[/]"
                )
                continue

            # Get all photos for this user
            photos = db_session.query(Photo).filter(Photo.user_id == user.id).all()
            if not photos:
                console.print(f"[yellow]No photos found for user {user.source_name}[/]")
                continue

            # Extract immich_ids from photos
            immich_ids = [str(photo.immich_id) for photo in photos if photo.immich_id]

            if not immich_ids:
                console.print(
                    f"[yellow]No valid immich_ids found for user {user.source_name}[/]"
                )
                continue

            console.print(
                f"\nProcessing user [blue]#{user.id}[/]: [yellow]{user.source_name} <{user.immich_email}>[/][dim with {len(immich_ids)} assets]"
            )
            console.print(f"Immich User ID: {user.immich_user_id}")

            if dry_run:
                console.print(
                    "[yellow]Dry run: Would update asset ownership in batches[/]"
                )
                continue

            # Count how many assets actually need updating
            total_to_update = immich_client.count_assets_needing_owner_update(
                asset_ids=immich_ids, new_owner_id=str(user.immich_user_id)
            )

            if total_to_update == 0:
                console.print("[green]✓ All assets already have the correct owner[/]")
                continue

            console.print(
                f"[yellow]Found [blue]{total_to_update}[/] assets to update...[/]"
            )

            with Progress(
                SpinnerColumn(),
                "[progress.description]{task.description}",
                BarColumn(bar_width=30),
                "[progress.percentage]{task.percentage:>3.0f}%",
                "•",
                "[cyan]{task.completed}/{task.total} assets",
                console=console,
            ) as progress:
                task = progress.add_task(
                    f"[cyan]Updating assets for user {user.immich_user_id}[/]",
                    total=total_to_update,
                )
                logger.info(
                    f"Updating assets for user {user.immich_user_id} ({total_to_update} assets)"
                )

                # Process batches
                for i in range(0, len(immich_ids), batch_size):
                    batch = immich_ids[i : i + batch_size]
                    try:
                        updated = immich_client.update_asset_owner(
                            asset_ids=batch, new_owner_id=str(user.immich_user_id)
                        )
                        if (
                            updated > 0
                        ):  # Only update progress if something was actually updated
                            progress.update(
                                task,
                                advance=updated,
                                description=f"[cyan]Updating assets for user {user.immich_user_id}[/]",
                            )
                            if len(immich_ids) < 500:
                                time.sleep(
                                    0.07
                                )  # Small delay to allow progress to update
                    except Exception as e:
                        progress.console.print(
                            f"[red]Error updating batch {i // batch_size + 1}: {e}[/]"
                        )
                        continue

    except Exception as e:
        console.print(f"[red]Error during database operations: {e}[/]")
        return
    finally:
        if "db_session" in locals():
            db_session.close()

    if storage_migration:
        console.print(
            "\nStarting 'storage template migration' job, this can take some time ..."
        )
        started = immich.start_job(JobName.STORAGETEMPLATEMIGRATION)
        if not started:
            console.print(
                "[red]Failed to start 'storage template migration' job, please start it manually[/]"
            )
            sys.exit(1)

        with Progress(
            SpinnerColumn(),
            "[progress.description]{task.description}",
            "•",
            "[elapsed]Elapsed: {task.fields[elapsed_time]}",
            console=console,
        ) as progress:
            start_time = time.time()
            task = progress.add_task(
                "[cyan]Running storage template migration...", elapsed_time="0:00:00"
            )

            job_status = immich.get_job_status(JobName.STORAGETEMPLATEMIGRATION)
            while job_status.job_counts.active > 0:
                elapsed = time.time() - start_time
                elapsed_str = time.strftime("%H:%M:%S", time.gmtime(elapsed))
                progress.update(
                    task,
                    description="[cyan]Running storage template migration...",
                    elapsed_time=elapsed_str,
                    refresh=True,
                )
                time.sleep(1)
                job_status = immich.get_job_status(JobName.STORAGETEMPLATEMIGRATION)

            progress.update(
                task,
                description="[green]Storage template migration complete!",
                elapsed_time=time.strftime(
                    "%H:%M:%S", time.gmtime(time.time() - start_time)
                ),
                refresh=True,
            )



import datetime
from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.asset_order import AssetOrder
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.album_user_response_dto import AlbumUserResponseDto
    from ..models.asset_response_dto import AssetResponseDto
    from ..models.user_response_dto import UserResponseDto


T = TypeVar("T", bound="AlbumResponseDto")


@_attrs_define
class AlbumResponseDto:
    """
    Attributes:
        album_name (str):
        album_thumbnail_asset_id (Union[None, str]):
        album_users (list['AlbumUserResponseDto']):
        asset_count (int):
        assets (list['AssetResponseDto']):
        created_at (datetime.datetime):
        description (str):
        has_shared_link (bool):
        id (str):
        is_activity_enabled (bool):
        owner (UserResponseDto):
        owner_id (str):
        shared (bool):
        updated_at (datetime.datetime):
        end_date (Union[Unset, datetime.datetime]):
        last_modified_asset_timestamp (Union[Unset, datetime.datetime]):
        order (Union[Unset, AssetOrder]):
        start_date (Union[Unset, datetime.datetime]):
    """

    album_name: str
    album_thumbnail_asset_id: Union[None, str]
    album_users: list["AlbumUserResponseDto"]
    asset_count: int
    assets: list["AssetResponseDto"]
    created_at: datetime.datetime
    description: str
    has_shared_link: bool
    id: str
    is_activity_enabled: bool
    owner: "UserResponseDto"
    owner_id: str
    shared: bool
    updated_at: datetime.datetime
    end_date: Union[Unset, datetime.datetime] = UNSET
    last_modified_asset_timestamp: Union[Unset, datetime.datetime] = UNSET
    order: Union[Unset, AssetOrder] = UNSET
    start_date: Union[Unset, datetime.datetime] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        album_name = self.album_name

        album_thumbnail_asset_id: Union[None, str]
        album_thumbnail_asset_id = self.album_thumbnail_asset_id

        album_users = []
        for album_users_item_data in self.album_users:
            album_users_item = album_users_item_data.to_dict()
            album_users.append(album_users_item)

        asset_count = self.asset_count

        assets = []
        for assets_item_data in self.assets:
            assets_item = assets_item_data.to_dict()
            assets.append(assets_item)

        created_at = self.created_at.isoformat()

        description = self.description

        has_shared_link = self.has_shared_link

        id = self.id

        is_activity_enabled = self.is_activity_enabled

        owner = self.owner.to_dict()

        owner_id = self.owner_id

        shared = self.shared

        updated_at = self.updated_at.isoformat()

        end_date: Union[Unset, str] = UNSET
        if not isinstance(self.end_date, Unset):
            end_date = self.end_date.isoformat()

        last_modified_asset_timestamp: Union[Unset, str] = UNSET
        if not isinstance(self.last_modified_asset_timestamp, Unset):
            last_modified_asset_timestamp = (
                self.last_modified_asset_timestamp.isoformat()
            )

        order: Union[Unset, str] = UNSET
        if not isinstance(self.order, Unset):
            order = self.order.value

        start_date: Union[Unset, str] = UNSET
        if not isinstance(self.start_date, Unset):
            start_date = self.start_date.isoformat()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "albumName": album_name,
                "albumThumbnailAssetId": album_thumbnail_asset_id,
                "albumUsers": album_users,
                "assetCount": asset_count,
                "assets": assets,
                "createdAt": created_at,
                "description": description,
                "hasSharedLink": has_shared_link,
                "id": id,
                "isActivityEnabled": is_activity_enabled,
                "owner": owner,
                "ownerId": owner_id,
                "shared": shared,
                "updatedAt": updated_at,
            }
        )
        if end_date is not UNSET:
            field_dict["endDate"] = end_date
        if last_modified_asset_timestamp is not UNSET:
            field_dict["lastModifiedAssetTimestamp"] = last_modified_asset_timestamp
        if order is not UNSET:
            field_dict["order"] = order
        if start_date is not UNSET:
            field_dict["startDate"] = start_date

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.album_user_response_dto import AlbumUserResponseDto
        from ..models.asset_response_dto import AssetResponseDto
        from ..models.user_response_dto import UserResponseDto

        d = dict(src_dict)
        album_name = d.pop("albumName")

        def _parse_album_thumbnail_asset_id(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        album_thumbnail_asset_id = _parse_album_thumbnail_asset_id(
            d.pop("albumThumbnailAssetId")
        )

        album_users = []
        _album_users = d.pop("albumUsers")
        for album_users_item_data in _album_users:
            album_users_item = AlbumUserResponseDto.from_dict(album_users_item_data)

            album_users.append(album_users_item)

        asset_count = d.pop("assetCount")

        assets = []
        _assets = d.pop("assets")
        for assets_item_data in _assets:
            assets_item = AssetResponseDto.from_dict(assets_item_data)

            assets.append(assets_item)

        created_at = isoparse(d.pop("createdAt"))

        description = d.pop("description")

        has_shared_link = d.pop("hasSharedLink")

        id = d.pop("id")

        is_activity_enabled = d.pop("isActivityEnabled")

        owner = UserResponseDto.from_dict(d.pop("owner"))

        owner_id = d.pop("ownerId")

        shared = d.pop("shared")

        updated_at = isoparse(d.pop("updatedAt"))

        _end_date = d.pop("endDate", UNSET)
        end_date: Union[Unset, datetime.datetime]
        if isinstance(_end_date, Unset):
            end_date = UNSET
        else:
            end_date = isoparse(_end_date)

        _last_modified_asset_timestamp = d.pop("lastModifiedAssetTimestamp", UNSET)
        last_modified_asset_timestamp: Union[Unset, datetime.datetime]
        if isinstance(_last_modified_asset_timestamp, Unset):
            last_modified_asset_timestamp = UNSET
        else:
            last_modified_asset_timestamp = isoparse(_last_modified_asset_timestamp)

        _order = d.pop("order", UNSET)
        order: Union[Unset, AssetOrder]
        if isinstance(_order, Unset):
            order = UNSET
        else:
            order = AssetOrder(_order)

        _start_date = d.pop("startDate", UNSET)
        start_date: Union[Unset, datetime.datetime]
        if isinstance(_start_date, Unset):
            start_date = UNSET
        else:
            start_date = isoparse(_start_date)

        album_response_dto = cls(
            album_name=album_name,
            album_thumbnail_asset_id=album_thumbnail_asset_id,
            album_users=album_users,
            asset_count=asset_count,
            assets=assets,
            created_at=created_at,
            description=description,
            has_shared_link=has_shared_link,
            id=id,
            is_activity_enabled=is_activity_enabled,
            owner=owner,
            owner_id=owner_id,
            shared=shared,
            updated_at=updated_at,
            end_date=end_date,
            last_modified_asset_timestamp=last_modified_asset_timestamp,
            order=order,
            start_date=start_date,
        )

        album_response_dto.additional_properties = d
        return album_response_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    cast,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.asset_response_dto import AssetResponseDto


T = TypeVar("T", bound="AssetDeltaSyncResponseDto")


@_attrs_define
class AssetDeltaSyncResponseDto:
    """
    Attributes:
        deleted (list[str]):
        needs_full_sync (bool):
        upserted (list['AssetResponseDto']):
    """

    deleted: list[str]
    needs_full_sync: bool
    upserted: list["AssetResponseDto"]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        deleted = self.deleted

        needs_full_sync = self.needs_full_sync

        upserted = []
        for upserted_item_data in self.upserted:
            upserted_item = upserted_item_data.to_dict()
            upserted.append(upserted_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "deleted": deleted,
                "needsFullSync": needs_full_sync,
                "upserted": upserted,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.asset_response_dto import AssetResponseDto

        d = dict(src_dict)
        deleted = cast(list[str], d.pop("deleted"))

        needs_full_sync = d.pop("needsFullSync")

        upserted = []
        _upserted = d.pop("upserted")
        for upserted_item_data in _upserted:
            upserted_item = AssetResponseDto.from_dict(upserted_item_data)

            upserted.append(upserted_item)

        asset_delta_sync_response_dto = cls(
            deleted=deleted,
            needs_full_sync=needs_full_sync,
            upserted=upserted,
        )

        asset_delta_sync_response_dto.additional_properties = d
        return asset_delta_sync_response_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="SessionCreateResponseDto")


@_attrs_define
class SessionCreateResponseDto:
    """
    Attributes:
        created_at (str):
        current (bool):
        device_os (str):
        device_type (str):
        id (str):
        is_pending_sync_reset (bool):
        token (str):
        updated_at (str):
        expires_at (Union[Unset, str]):
    """

    created_at: str
    current: bool
    device_os: str
    device_type: str
    id: str
    is_pending_sync_reset: bool
    token: str
    updated_at: str
    expires_at: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        created_at = self.created_at

        current = self.current

        device_os = self.device_os

        device_type = self.device_type

        id = self.id

        is_pending_sync_reset = self.is_pending_sync_reset

        token = self.token

        updated_at = self.updated_at

        expires_at = self.expires_at

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "createdAt": created_at,
                "current": current,
                "deviceOS": device_os,
                "deviceType": device_type,
                "id": id,
                "isPendingSyncReset": is_pending_sync_reset,
                "token": token,
                "updatedAt": updated_at,
            }
        )
        if expires_at is not UNSET:
            field_dict["expiresAt"] = expires_at

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        created_at = d.pop("createdAt")

        current = d.pop("current")

        device_os = d.pop("deviceOS")

        device_type = d.pop("deviceType")

        id = d.pop("id")

        is_pending_sync_reset = d.pop("isPendingSyncReset")

        token = d.pop("token")

        updated_at = d.pop("updatedAt")

        expires_at = d.pop("expiresAt", UNSET)

        session_create_response_dto = cls(
            created_at=created_at,
            current=current,
            device_os=device_os,
            device_type=device_type,
            id=id,
            is_pending_sync_reset=is_pending_sync_reset,
            token=token,
            updated_at=updated_at,
            expires_at=expires_at,
        )

        session_create_response_dto.additional_properties = d
        return session_create_response_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import (
    Any,
    TypeVar,
    Union,
)
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.asset_order import AssetOrder
from ..types import UNSET, Unset

T = TypeVar("T", bound="UpdateAlbumDto")


@_attrs_define
class UpdateAlbumDto:
    """
    Attributes:
        album_name (Union[Unset, str]):
        album_thumbnail_asset_id (Union[Unset, UUID]):
        description (Union[Unset, str]):
        is_activity_enabled (Union[Unset, bool]):
        order (Union[Unset, AssetOrder]):
    """

    album_name: Union[Unset, str] = UNSET
    album_thumbnail_asset_id: Union[Unset, UUID] = UNSET
    description: Union[Unset, str] = UNSET
    is_activity_enabled: Union[Unset, bool] = UNSET
    order: Union[Unset, AssetOrder] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        album_name = self.album_name

        album_thumbnail_asset_id: Union[Unset, str] = UNSET
        if not isinstance(self.album_thumbnail_asset_id, Unset):
            album_thumbnail_asset_id = str(self.album_thumbnail_asset_id)

        description = self.description

        is_activity_enabled = self.is_activity_enabled

        order: Union[Unset, str] = UNSET
        if not isinstance(self.order, Unset):
            order = self.order.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if album_name is not UNSET:
            field_dict["albumName"] = album_name
        if album_thumbnail_asset_id is not UNSET:
            field_dict["albumThumbnailAssetId"] = album_thumbnail_asset_id
        if description is not UNSET:
            field_dict["description"] = description
        if is_activity_enabled is not UNSET:
            field_dict["isActivityEnabled"] = is_activity_enabled
        if order is not UNSET:
            field_dict["order"] = order

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        album_name = d.pop("albumName", UNSET)

        _album_thumbnail_asset_id = d.pop("albumThumbnailAssetId", UNSET)
        album_thumbnail_asset_id: Union[Unset, UUID]
        if isinstance(_album_thumbnail_asset_id, Unset):
            album_thumbnail_asset_id = UNSET
        else:
            album_thumbnail_asset_id = UUID(_album_thumbnail_asset_id)

        description = d.pop("description", UNSET)

        is_activity_enabled = d.pop("isActivityEnabled", UNSET)

        _order = d.pop("order", UNSET)
        order: Union[Unset, AssetOrder]
        if isinstance(_order, Unset):
            order = UNSET
        else:
            order = AssetOrder(_order)

        update_album_dto = cls(
            album_name=album_name,
            album_thumbnail_asset_id=album_thumbnail_asset_id,
            description=description,
            is_activity_enabled=is_activity_enabled,
            order=order,
        )

        update_album_dto.additional_properties = d
        return update_album_dto

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Shared database models for immichporter."""

from sqlalchemy import (
    create_engine,
    Integer,
    String,
    Boolean,
    DateTime,
    ForeignKey,
    UniqueConstraint,
    func,
)
from sqlalchemy.orm import (
    declarative_base,
    sessionmaker,
    relationship,
    Mapped,
    mapped_column,
)

# Configuration
DATABASE_PATH = "immichporter.db"
BRAVE_EXECUTABLE = "/usr/bin/brave-browser"

# SQLAlchemy setup
Base = declarative_base()
engine = create_engine(f"sqlite:///{DATABASE_PATH}")
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


class Album(Base):
    """SQLAlchemy model for albums."""

    __tablename__ = "albums"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    source_title: Mapped[str] = mapped_column(String, unique=False, nullable=False)
    source_type: Mapped[str] = mapped_column(
        String, nullable=False
    )  # 'gphoto', 'local', etc.
    immich_title: Mapped[str | None] = mapped_column(
        String, unique=False, nullable=True
    )
    immich_id: Mapped[str | None] = mapped_column(
        String, nullable=True, unique=True, index=True
    )
    items: Mapped[int | None] = mapped_column(Integer)
    processed_items: Mapped[int] = mapped_column(Integer, default=0)
    shared: Mapped[bool] = mapped_column(Boolean, default=False)
    created_at: Mapped[DateTime] = mapped_column(DateTime, default=func.now())
    updated_at: Mapped[DateTime] = mapped_column(
        DateTime, default=func.now(), onupdate=func.now()
    )
    source_url: Mapped[str] = mapped_column(String, unique=True, nullable=False)

    # Relationships
    photos: Mapped[list["Photo"]] = relationship(
        "Photo", back_populates="album", cascade="all, delete-orphan"
    )
    errors: Mapped[list["Error"]] = relationship("Error", back_populates="album")
    users: Mapped[list["User"]] = relationship(
        "User", secondary="album_users", back_populates="albums"
    )

    def __repr__(self):
        return f"<Album(source_title='{self.source_title}', source_type='{self.source_type}', items={self.items}, shared={self.shared})>"


class User(Base):
    """SQLAlchemy model for users."""

    __tablename__ = "users"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    source_name: Mapped[str] = mapped_column(String, unique=False, nullable=False)
    source_type: Mapped[str] = mapped_column(
        String, nullable=False
    )  # 'gphoto', 'local', etc.
    immich_name: Mapped[str | None] = mapped_column(String, unique=False, nullable=True)
    immich_email: Mapped[str | None] = mapped_column(String, nullable=True)
    immich_user_id: Mapped[str | None] = mapped_column(String, nullable=True)
    immich_initial_password: Mapped[str | None] = mapped_column(String, nullable=True)
    add_to_immich: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)
    created_at: Mapped[DateTime] = mapped_column(DateTime, default=func.now())
    albums: Mapped[list["Album"]] = relationship(
        "Album", secondary="album_users", back_populates="users"
    )
    photos: Mapped[list["Photo"]] = relationship("Photo", back_populates="user")

    def __repr__(self):
        return f"<User(source_name='{self.source_name}', source_type='{self.source_type}', immich_name='{self.immich_name}', immich_email='{self.immich_email}')>"


class Photo(Base):
    """SQLAlchemy model for photos."""

    __tablename__ = "photos"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    filename: Mapped[str] = mapped_column(String, nullable=False)
    date_taken: Mapped[DateTime | None] = mapped_column(DateTime)
    album_id: Mapped[int | None] = mapped_column(
        Integer, ForeignKey("albums.id", ondelete="CASCADE")
    )
    user_id: Mapped[int | None] = mapped_column(
        Integer, ForeignKey("users.id", ondelete="SET NULL")
    )
    created_at: Mapped[DateTime] = mapped_column(DateTime, default=func.now())
    source_id: Mapped[str] = mapped_column(
        String, nullable=False, unique=True, index=True
    )
    immich_id: Mapped[str | None] = mapped_column(
        String, nullable=True, unique=True, index=True
    )
    saved_to_your_photos: Mapped[bool] = mapped_column(
        Boolean, default=False, nullable=False
    )

    # Relationships
    album: Mapped[Album | None] = relationship("Album", back_populates="photos")
    user: Mapped[User | None] = relationship("User", back_populates="photos")

    def __repr__(self):
        return f"<Photo(filename='{self.filename}', date_taken={self.date_taken})>"


class Error(Base):
    """SQLAlchemy model for errors."""

    __tablename__ = "errors"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    error_message: Mapped[str] = mapped_column(String, nullable=False)
    album_id: Mapped[int | None] = mapped_column(
        Integer, ForeignKey("albums.id", ondelete="SET NULL")
    )
    created_at: Mapped[DateTime] = mapped_column(DateTime, default=func.now())

    # Relationships
    album: Mapped[Album | None] = relationship("Album", back_populates="errors")

    def __repr__(self):
        return f"<Error(error_message='{self.error_message[:50]}...')>"


class AlbumUser(Base):
    """SQLAlchemy model for album-user relationships."""

    __tablename__ = "album_users"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    album_id: Mapped[int] = mapped_column(
        Integer, ForeignKey("albums.id", ondelete="CASCADE")
    )
    user_id: Mapped[int] = mapped_column(
        Integer, ForeignKey("users.id", ondelete="CASCADE")
    )
    created_at: Mapped[DateTime] = mapped_column(DateTime, default=func.now())
    __table_args__ = (
        UniqueConstraint("album_id", "user_id", name="unique_album_user"),
    )

    def __repr__(self):
        return f"<AlbumUser(album_id={self.album_id}, user_id={self.user_id})>"
